import wx
from functools import cmp_to_key

# import wx.lib.agw.ribbon as RB
import wx.lib.agw.ribbon as RB
from wx.lib.agw.ribbon import art
from wx.lib.agw.ribbon.art import *

from pytigon_gui.guilib.events import *
from pytigon_gui.toolbar.basetoolbar import (
    BaseHtmlPanel,
    ToolbarBar,
    ToolbarPage,
    ToolbarPanel,
    ToolbarButton,
)

_ = wx.GetTranslation


MSW_STYLE = True
ORG_LIKE_PRIMARY = None


def like_primary(primary_hsl, h, s, l, x=None):
    if x != None:
        c1 = ORG_LIKE_PRIMARY(primary_hsl, h, 0, l * 1.5, x)
    else:
        c1 = ORG_LIKE_PRIMARY(primary_hsl, h, 0, l * 1.5)
    c2 = wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE)
    r2 = c2.Red()
    g2 = c2.Green()
    b2 = c2.Blue()
    if r2 == 0:
        r2 = 1
    if g2 == 0:
        g2 = 1
    if b2 == 0:
        b2 = 1
    m = ((1.0 * c1.Red()) / r2 + (1.0 * c1.Green()) / g2 + (1.0 * c1.Blue()) / b2) / 3
    x1 = int(c2.Red() * m)
    x2 = int(c2.Green() * m)
    x3 = int(c2.Blue() * m)
    if x1 > 255:
        x1 = 255
    if x2 > 255:
        x2 = 255
    if x3 > 255:
        x3 = 255
    c3 = wx.Colour(x1, x2, x3)
    return c3


class ModernHtmlPanel(BaseHtmlPanel):
    def get_width(self):
        return self.page.parent_bar.get_bar_width()

    def get_height(self):
        return self.page.parent_bar.get_bar_height()

    def set_page(self, html_page):
        super().set_page(html_page)
        self.page.parent_bar.update()
        self.page.parent_bar.SetActivePage(self.page)


class ModernToolbarButton(ToolbarButton):
    def __init__(
        self,
        parent_panel,
        id,
        title,
        bitmap,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        ToolbarButton.__init__(
            self, parent_panel, id, title, bitmap, bitmap_disabled, kind
        )


class ModernToolbarPanel(ToolbarPanel, RB.RibbonPanel):
    def __init__(self, parent_page, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        RB.RibbonPanel.__init__(
            self,
            parent_page,
            wx.ID_ANY,
            title,
            wx.NullBitmap,
            wx.DefaultPosition,
            wx.DefaultSize,
            RB.RIBBON_PANEL_NO_AUTO_MINIMISE,
        )
        ToolbarPanel.__init__(self, parent_page, title, kind)

        if self.kind == ToolbarPanel.TYPE_PANEL_TOOLBAR:
            self.toolbar = RB.RibbonToolBar(self)
            # self.toolbar = SchRibbonToolBar(self)
        elif self.kind == ToolbarPanel.TYPE_PANEL_BUTTONBAR:
            self.toolbar = RB.RibbonButtonBar(self)

        self.lock = False

    def OnInternalIdle(self):
        if not self.lock:
            self.lock = True
            if self.toolbar:
                self.toolbar.UpdateWindowUI(wx.UPDATE_UI_FROMIDLE)
            self.lock = False

    def Append(self, b):
        if self.kind == ToolbarPanel.TYPE_PANEL_TOOLBAR:
            if b.kind == ToolbarButton.TYPE_SIMPLE:
                self.toolbar.AddSimpleTool(b.id, b.bitmap, b.title)
            elif b.kind == ToolbarButton.TYPE_DROPDOWN:
                self.toolbar.AddDropdownTool(b.id, b.bitmap, b.title)
            elif b.kind == ToolbarButton.TYPE_HYBRID:
                self.toolbar.AddHybridTool(b.id, b.bitmap, b.title)
            elif b.kind == ToolbarButton.TYPE_TOOGLE:
                self.toolbar.AddToggleTool(b.id, b.bitmap, b.title)
            elif b.kind == ToolbarButton.TYPE_PANEL:
                pass
            elif b.kind == ToolbarButton.TYPE_SEPARATOR:
                self.toolbar.AddSeparator()
        elif self.kind == ToolbarPanel.TYPE_PANEL_BUTTONBAR:
            if b.kind == ToolbarButton.TYPE_SIMPLE:
                self.toolbar.AddSimpleButton(b.id, b.title, b.bitmap, "")
            elif b.kind == ToolbarButton.TYPE_DROPDOWN:
                self.toolbar.AddDropdownButton(b.id, b.title, b.bitmap, "")
            elif b.kind == ToolbarButton.TYPE_HYBRID:
                self.toolbar.AddHybridButton(b.id, b.title, b.bitmap, "")
            elif b.kind == ToolbarButton.TYPE_TOOGLE:
                self.toolbar.AddToggleButton(b.id, b.title, b.bitmap, "")
            elif b.kind == ToolbarButton.TYPE_PANEL:
                pass
            # elif b.kind == ToolbarButton.TYPE_SEPARATOR:
            #    self.toolbar.AddSeparator()
        else:
            pass

    def create_button(
        self,
        id,
        title,
        bitmap=None,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        b = ModernToolbarButton(self, id, title, bitmap, bitmap_disabled, kind)
        self.Append(b)
        return b

    def add_separator(self):
        b = ModernToolbarButton(
            self, 0, "", None, None, kind=ToolbarButton.TYPE_SEPARATOR
        )
        self.Append(b)
        return b


class ModernToolbarPage(ToolbarPage, RB.RibbonPage):
    def __init__(self, parent_bar, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        ToolbarPage.__init__(self, parent_bar, title, kind)
        RB.RibbonPage.__init__(self, parent_bar, wx.ID_ANY, self.title)

    def create_panel(self, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        return ModernToolbarPanel(self, title, kind)

    def create_html_panel(self, title):
        p = self.create_panel(title)
        return ModernHtmlPanel(self, p)


class ModernToolbarBar(ToolbarBar, RB.RibbonBar):
    def __init__(self, parent, gui_style):
        RB.RibbonBar.__init__(self, parent, wx.ID_ANY)
        ToolbarBar.__init__(self, parent, gui_style)

    def create_page(self, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        return ModernToolbarPage(self, title, kind)

    def create(self):
        self.Realize()

        def _realize():
            for child in self.Children:
                child.Realize()

        wx.CallAfter(_realize)

    def close(self):
        self.Close()

    def update(self):
        size = self.GetSize()
        self.SetSize(wx.Size(size.GetWidth() - 1, size.GetHeight()))
        self.SetSize(wx.Size(size.GetWidth(), size.GetHeight()))

    def bind_ui(self, fun, id=wx.ID_ANY):
        self.Bind(wx.EVT_UPDATE_UI, fun, id=id)

    def bind(self, fun, id=wx.ID_ANY, e=None):
        if e:
            self.Bind(e, fun, id=id)
        else:
            self.Bind(RB.EVT_RIBBONBUTTONBAR_CLICKED, fun, id=id)
            self.Bind(RB.EVT_RIBBONTOOLBAR_CLICKED, fun, id=id)

    def bind_dropdown(self, fun, id):
        self.bar.Bind(RB.EVT_RIBBONTOOLBAR_DROPDOWN_CLICKED, fun, id=id)

    def un_bind(self, id, e=None):
        if e:
            self.Unbind(e, id=id)
        else:
            self.Unbind(RB.EVT_RIBBONBUTTONBAR_CLICKED, id=id)

    def get_bar_height(self):
        s = (48, 48)
        ret = self.GetArtProvider().GetButtonBarButtonSize(
            self,
            self,
            art.RIBBON_BUTTON_NORMAL,
            art.RIBBON_BUTTONBAR_BUTTON_LARGE,
            "TXT",
            s,
            s,
        )
        size_ret = ret[1]
        return size_ret.GetHeight()

    def get_bar_width(self):
        return 2000

    def remove_page(self, title):
        self.SetActivePage(0)
        for page_info in self._pages:
            if page_info.page.title == title:
                self._pages.remove(page_info)
        self.Update()
        super().remove_page(title)

    def Realize(self):
        if MSW_STYLE:
            global ORG_LIKE_PRIMARY
            if not ORG_LIKE_PRIMARY:
                ORG_LIKE_PRIMARY = RB.art_msw.LikePrimary
                RB.art_msw.LikePrimary = like_primary
            provider = RB.RibbonMSWArtProvider()
            (dummy, secondary, tertiary) = provider.GetColourScheme(None, 1, 1)
            colour = wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE)
            colour2 = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNTEXT)
            provider.SetColourScheme(colour, secondary, colour2)
            provider._tab_label_colour = colour2
            provider._button_bar_label_colour = colour2
        else:
            provider = RB.RibbonAUIArtProvider()
        self.SetArtProvider(provider)
        RB.RibbonBar.Realize(self)

    def RecalculateTabSizes(self):
        """Recalculates the :class:`RibbonBar` tab sizes."""

        numtabs = len(self._pages)

        if numtabs == 0:
            return

        width = (
            self.GetSize().GetWidth() - self._tab_margin_left - self._tab_margin_right
        )
        tabsep = self._art.GetMetric(RIBBON_ART_TAB_SEPARATION_SIZE)
        x = self._tab_margin_left
        y = 0

        if width >= self._tabs_total_width_ideal:
            # Simple case: everything at ideal width
            for info in self._pages:
                info.rect.x = x
                info.rect.y = y
                info.rect.width = info.ideal_width
                info.rect.height = self._tab_height
                x += info.rect.width + tabsep

            self._tab_scroll_buttons_shown = False
            self._tab_scroll_left_button_rect.SetWidth(0)
            self._tab_scroll_right_button_rect.SetWidth(0)

        elif width < self._tabs_total_width_minimum:
            # Simple case: everything minimum with scrollbar
            for info in self._pages:
                info.rect.x = x
                info.rect.y = y
                info.rect.width = info.minimum_width
                info.rect.height = self._tab_height
                x += info.rect.width + tabsep

            if not self._tab_scroll_buttons_shown:
                self._tab_scroll_left_button_state = RIBBON_SCROLL_BTN_NORMAL
                self._tab_scroll_right_button_state = RIBBON_SCROLL_BTN_NORMAL
                self._tab_scroll_buttons_shown = True

            temp_dc = wx.ClientDC(self)
            self._tab_scroll_left_button_rect.SetWidth(
                self._art.GetScrollButtonMinimumSize(
                    temp_dc,
                    self,
                    RIBBON_SCROLL_BTN_LEFT
                    | RIBBON_SCROLL_BTN_NORMAL
                    | RIBBON_SCROLL_BTN_FOR_TABS,
                ).GetWidth()
            )
            self._tab_scroll_left_button_rect.SetHeight(self._tab_height)
            self._tab_scroll_left_button_rect.SetX(self._tab_margin_left)
            self._tab_scroll_left_button_rect.SetY(0)
            self._tab_scroll_right_button_rect.SetWidth(
                self._art.GetScrollButtonMinimumSize(
                    temp_dc,
                    self,
                    RIBBON_SCROLL_BTN_RIGHT
                    | RIBBON_SCROLL_BTN_NORMAL
                    | RIBBON_SCROLL_BTN_FOR_TABS,
                ).GetWidth()
            )
            self._tab_scroll_right_button_rect.SetHeight(self._tab_height)
            self._tab_scroll_right_button_rect.SetX(
                self.GetClientSize().GetWidth()
                - self._tab_margin_right
                - self._tab_scroll_right_button_rect.GetWidth()
            )
            self._tab_scroll_right_button_rect.SetY(0)

            if self._tab_scroll_amount == 0:
                self._tab_scroll_left_button_rect.SetWidth(0)

            elif self._tab_scroll_amount + width >= self._tabs_total_width_minimum:
                self._tab_scroll_amount = self._tabs_total_width_minimum - width
                self._tab_scroll_right_button_rect.SetX(
                    self._tab_scroll_right_button_rect.GetX()
                    + self._tab_scroll_right_button_rect.GetWidth()
                )
                self._tab_scroll_right_button_rect.SetWidth(0)

            for info in self._pages:
                info.rect.x -= self._tab_scroll_amount

        else:
            self._tab_scroll_buttons_shown = False
            self._tab_scroll_left_button_rect.SetWidth(0)
            self._tab_scroll_right_button_rect.SetWidth(0)
            # Complex case: everything sized such that: minimum <= width < ideal
            #
            #   Strategy:
            #     1) Uniformly reduce all tab widths from ideal to small_must_have_separator_width
            #     2) Reduce the largest tab by 1 pixel, repeating until all tabs are same width (or at minimum)
            #     3) Uniformly reduce all tabs down to their minimum width
            #
            smallest_tab_width = 10000
            total_small_width = tabsep * (numtabs - 1)

            for info in self._pages:
                if info.small_must_have_separator_width < smallest_tab_width:
                    smallest_tab_width = info.small_must_have_separator_width

                total_small_width += info.small_must_have_separator_width

            if width >= total_small_width:
                # Do (1)
                total_delta = self._tabs_total_width_ideal - total_small_width
                total_small_width -= tabsep * (numtabs - 1)
                width -= tabsep * (numtabs - 1)
                for info in self._pages:
                    delta = info.ideal_width - info.small_must_have_separator_width
                    info.rect.x = x
                    info.rect.y = y
                    info.rect.width = (
                        info.small_must_have_separator_width
                        + delta * (width - total_small_width) // total_delta
                    )
                    info.rect.height = self._tab_height

                    x += info.rect.width + tabsep
                    total_delta -= delta
                    total_small_width -= info.small_must_have_separator_width
                    width -= info.rect.width

            else:
                total_small_width = tabsep * (numtabs - 1)
                for info in self._pages:
                    if info.minimum_width < smallest_tab_width:
                        total_small_width += smallest_tab_width
                    else:
                        total_small_width += info.minimum_width

                if width >= total_small_width:
                    # Do (2)
                    sorted_pages = []
                    for info in self._pages:
                        # Sneaky obj array trickery to not copy the tab descriptors
                        sorted_pages.append(info)

                    sorted_pages.sort(
                        key=cmp_to_key(self.OrderPageTabInfoBySmallWidthAsc)
                    )
                    width -= tabsep * (numtabs - 1)

                    for i, info in enumerate(self._pages):
                        if (
                            info.small_must_have_separator_width * (numtabs - i)
                            <= width
                        ):
                            info.rect.width = info.small_must_have_separator_width
                        else:
                            info.rect.width = width // (numtabs - i)

                        width -= info.rect.width

                    for i, info in enumerate(self._pages):
                        info.rect.x = x
                        info.rect.y = y
                        info.rect.height = self._tab_height
                        x += info.rect.width + tabsep
                        sorted_pages.pop(numtabs - (i + 1))

                else:
                    # Do (3)
                    total_small_width = (smallest_tab_width + tabsep) * numtabs - tabsep
                    total_delta = total_small_width - self._tabs_total_width_minimum
                    total_small_width = self._tabs_total_width_minimum - tabsep * (
                        numtabs - 1
                    )
                    width -= tabsep * (numtabs - 1)

                    for info in self._pages:
                        delta = smallest_tab_width - info.minimum_width
                        info.rect.x = x
                        info.rect.y = y
                        info.rect.width = (
                            info.minimum_width
                            + delta * (width - total_small_width) // total_delta
                        )
                        info.rect.height = self._tab_height

                        x += info.rect.width + tabsep
                        total_delta -= delta
                        total_small_width -= info.minimum_width
                        width -= info.rect.width



import wx
import wx.grid
import csv
import io

from wx.grid import (
    GridCellAttr,
    GridTableMessage,
    GRIDTABLE_NOTIFY_ROWS_DELETED,
    GRIDTABLE_NOTIFY_ROWS_APPENDED,
)

from pytigon_gui.guictrl.grid import popupcelleditors
from pytigon_gui.guictrl.grid.renderers import (
    ExtStringRenderer,
    IconAndStringRenderer,
    DateTimeRenderer,
)
from pytigon_lib.schtools.schjson import json_dumps, json_loads

_ = wx.GetTranslation


class SchTableGrid(wx.grid.Grid):
    SORT_COLOR = "ORANGE"

    GET_ID = 0
    VIEW = 1
    EDIT = 2

    def __init__(
        self,
        table,
        address,
        parent,
        typ=EDIT,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.WANTS_CHARS | wx.TAB_TRAVERSAL,
        name=wx.PanelNameStr,
    ):
        self.address = address
        wx.grid.Grid.__init__(self, parent, wx.ID_ANY, pos, size, style, name)

        self.RegisterDataType(
            "s", IconAndStringRenderer(), wx.grid.GridCellTextEditor()
        )
        self.RegisterDataType(
            "x", ExtStringRenderer(), popupcelleditors.GenericPopupCellEditor()
        )
        self.RegisterDataType(
            "y", ExtStringRenderer(), popupcelleditors.ListPopupCellEditor()
        )
        self.RegisterDataType(
            "f", ExtStringRenderer(), popupcelleditors.GenericPopupCellEditor()
        )
        self.RegisterDataType(
            "str", wx.grid.GridCellStringRenderer(), wx.grid.GridCellTextEditor()
        )
        self.RegisterDataType(
            "string", wx.grid.GridCellStringRenderer(), wx.grid.GridCellTextEditor()
        )
        self.RegisterDataType(
            "datetime", DateTimeRenderer(), wx.grid.GridCellTextEditor()
        )
        self.RegisterDataType(
            "date", DateTimeRenderer(), popupcelleditors.DatePopupDataCellEditor()
        )
        self.RegisterDataType(
            "int", ExtStringRenderer(), popupcelleditors.GenericPopupCellEditor()
        )
        self.RegisterDataType(
            "bool", wx.grid.GridCellBoolRenderer(), wx.grid.GridCellBoolEditor()
        )

        if hasattr(table, "proxy"):
            self.SetTable(table, True)
            sizes = table.proxy.GetColSize()
            names = table.GetColNames()
            for i in range(len(sizes)):
                if 7 * len(str(names[i + 1])) > 7 * int(sizes[i]):
                    l = len(str(names[i + 1]))
                    self.SetColSize(i, 7 * l)
                else:
                    l = int(sizes[i])
                    self.SetColSize(i, 7 * l)
        else:
            if table.auto_size == "short":
                self.SetTable(table, False)
                self.SetMargins(0, 0)
                # self.AutoSizeColumns(False)
                # self.AutoSizeRows(True)
                height = 5
                for row in range(0, table.per_page):
                    h = self.GetRowSize(row)
                    if h > height:
                        height = h
                width = []
                for col in range(0, self.GetNumberCols()):
                    width.append(self.GetColSize(col))
                table.enable(True)
                self.SetTable(table, True)
                self.SetDefaultRowSize(height, True)
                i = 0
                for w in width:
                    self.SetColSize(i, w)
                    i += 1
            else:
                table.enable(True)
                self.SetTable(table, True)
                # self.AutoSizeColumns(False)
                # self.AutoSizeRows(True)

            sizes = table.get_sizes()
            if sizes:
                self.set_col_width(sizes)

                # width = []
                # for col in range(0, self.GetNumberCols()):
                #    width.append(self.GetColSize(col))
                # self.SetDefaultRowSize(height, True)

        try:
            self.SetSelectionMode(wx.grid.Grid.SelectRows)
        except:
            self.SetSelectionMode(wx.grid.Grid.GridSelectRows)

        self.typ = typ
        self.readonly = False
        self.panel = None
        self.oldrow = -1
        self.default_command = "edit"
        self.last_action = ""

        self.SetRowLabelSize(1)
        self.SetSize((400, 300))
        self.SelectRow(0)
        self.SetScrollLineY(1)

        self.Bind(wx.grid.EVT_GRID_SELECT_CELL, self.on_select_cell)
        self.Bind(wx.grid.EVT_GRID_CELL_LEFT_CLICK, self.on_cell_left_click)
        self.Bind(wx.grid.EVT_GRID_CELL_LEFT_DCLICK, self.on_cell_left_dclick)
        self.Bind(wx.grid.EVT_GRID_RANGE_SELECT, self.on_range_selected)
        self.Bind(wx.grid.EVT_GRID_CMD_LABEL_LEFT_CLICK, self.on_label_left_click)
        self.Bind(wx.grid.EVT_GRID_CMD_LABEL_RIGHT_CLICK, self.on_label_right_click)
        self.Bind(wx.grid.EVT_GRID_CMD_LABEL_LEFT_DCLICK, self.on_label_left_dclick)
        self.Bind(wx.grid.EVT_GRID_CMD_CELL_RIGHT_CLICK, self.on_cell_right_click)
        self.Bind(wx.EVT_MENU, self.begin_edit)
        self.Bind(wx.grid.EVT_GRID_CMD_CELL_CHANGED, self.on_cell_change)
        self.Bind(wx.grid.EVT_GRID_CELL_CHANGING, self.on_cell_change)
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)

        column_label_window = self.GetGridColLabelWindow()
        self.Bind(wx.EVT_PAINT, self.on_column_header_paint, column_label_window)

        self.EnableDragCell(False)
        self.EnableDragGridSize(False)

    def begin_edit(self, event=None):
        if event is None:
            self.GetEventHandler().AddPendingEvent(wx.CommandEvent(wx.EVT_MENU.typeId))
        else:
            wx.CallAfter(self.EnableCellEditControl, enable=True)

    def set_typ(self, typ):
        self.typ = typ

    def get_min_size(self):
        return wx.Size(400, 300)

    def get_best_size(self):
        return wx.Size(400, 300)

    def set_col_width(self, sizes):
        ps = wx.SMALL_FONT.GetPointSize()
        i = 0
        for w in sizes[:-1]:
            if w > 64:
                w = 64
            if w < 4:
                w = 4
            self.SetColSize(i, int(w * ps))
            i += 1

    def set_panel(self, panel):
        self.panel = panel

    def get_html_parent(self):
        parent = self.GetParent()
        while parent != None:
            if parent.__class__.__name__ == "SchForm":
                return parent
            parent = parent.GetParent()
        return None

    def CanClose(self):
        if self.typ == self.VIEW or self.readonly:
            return True
        if self.GetTable().data_changed():
            msg = _("Data not saved, save now?")
            dlg = wx.MessageDialog(
                self,
                msg,
                _("Pytigon"),
                wx.YES_NO | wx.YES_DEFAULT | wx.CANCEL | wx.ICON_QUESTION,
            )
            ret = dlg.ShowModal()
            if ret == wx.ID_CANCEL:
                dlg.Destroy()
                return False
            elif ret == wx.ID_YES:
                self.GetTable().commit()
            dlg.Destroy()
            return True
        else:
            return True

    def goto_first_row(self):
        col = self.GetGridCursorCol()
        self.SetGridCursor(0, col)
        self.MakeCellVisible(0, col)

    def goto_last_row(self):
        col = self.GetGridCursorCol()
        self.SetGridCursor(self.GetNumberRows() - 1, col)
        self.MakeCellVisible(self.GetNumberRows() - 1, col)

    def goto_next_row(self):
        self.DisableCellEditControl()
        col = self.GetGridCursorCol()
        new_row = self.GetGridCursorRow() + 1
        if new_row < self.GetTable().GetNumberRows():
            self.SetGridCursor(new_row, col)
            self.MakeCellVisible(new_row, col)
            self.SelectRow(new_row)

    def goto_prev_row(self):
        self.DisableCellEditControl()
        col = self.GetGridCursorCol()
        new_row = self.GetGridCursorRow() - 1
        if new_row >= 0:
            self.SetGridCursor(new_row, col)
            self.MakeCellVisible(new_row, col)
            self.SelectRow(new_row)

    def on_cell_change(self, evt):
        row = evt.GetRow()
        col = evt.GetCol()
        name = self.GetTable().GetColNames()[evt.GetCol() + 1]
        autocols = self.GetTable().GetAutoCols()
        if name in autocols:
            value = self.GetCellValue(evt.GetRow(), evt.GetCol())
            names = self.GetTable().GetColNames()[1:]
            rec = []
            for i in range(len(names)):
                rec.append(self.GetTable().GetValue(row, i))
            rec2 = self.GetTable().auto_update(name, names, rec)
            for i in range(len(rec2) - 1):
                self.GetTable().SetValue(row, i, rec2[i])
        evt.Skip()

    def on_navigate(self, evt):
        self.DisableCellEditControl()
        forward = evt.GetDirection()
        if forward:
            success = self.MoveCursorRight(False)
            if not success:
                new_row = self.GetGridCursorRow() + 1
                if self.GetGridCursorRow() < self.GetTable().GetNumberRows():
                    self.SetGridCursor(new_row, 0)
                    self.MakeCellVisible(new_row, 0)
                else:
                    pass
        else:
            success = self.MoveCursorLeft(False)
            if not success:
                new_row = self.GetGridCursorRow() - 1
                if new_row >= 0:
                    self.SetGridCursor(new_row, self.GetTable().GetNumberCols() - 1)
                    self.MakeCellVisible(new_row, self.GetTable().GetNumberCols() - 1)
                else:
                    pass
        evt.Skip()

    def on_column_header_paint(self, evt):
        w = self.GetGridColLabelWindow()
        dc = wx.PaintDC(w)
        client_rect = w.GetClientRect()
        font = wx.NORMAL_FONT
        tot_col_size = -self.GetViewStart()[0] * self.GetScrollPixelsPerUnit()[0]
        fs = font.GetPointSize()
        fs2 = int(0.8 * fs)
        for col in range(self.GetNumberCols()):
            dc.SetBrush(wx.Brush(self.SORT_COLOR, wx.TRANSPARENT))
            dc.SetTextForeground(wx.BLACK)
            col_size = self.GetColSize(col)
            rect = (tot_col_size, 0, col_size, 32)
            dc.SetPen(wx.GREY_PEN)
            dc.DrawLine(rect[0], rect[1], rect[0] + rect[2], rect[1])
            dc.SetPen(wx.WHITE_PEN)
            dc.DrawLine(rect[0], rect[1] + 1, rect[0] + rect[2], rect[1] + 1)
            dc.DrawLine(rect[0], rect[1] + 1, rect[0], rect[1] + rect[3])
            dc.SetPen(wx.GREY_PEN)
            dc.DrawLine(
                rect[0],
                (rect[1] + rect[3]) - 1,
                rect[0] + rect[2],
                (rect[1] + rect[3]) - 1,
            )
            dc.DrawLine(
                (rect[0] + rect[2]) - 1,
                rect[1],
                (rect[0] + rect[2]) - 1,
                rect[1] + rect[3],
            )
            dc.SetPen(wx.BLACK_PEN)
            tot_col_size += col_size
            srt = self.GetTable().get_sort_nr(col)
            if srt != 0:
                font.SetWeight(wx.BOLD)
                left = rect[0] + 3
                top = rect[1] + 3
                dc.SetBrush(wx.Brush(self.SORT_COLOR, wx.SOLID))
                if srt > 0:
                    dc.DrawPolygon([(left, top), (left + 6, top), (left + 3, top + 4)])
                else:
                    dc.DrawPolygon(
                        [(left + 3, top), (left + 6, top + 4), (left, top + 4)]
                    )
                if srt < 0:
                    srt = srt * -1
                dc.SetFont(wx.SMALL_FONT)
                dc.DrawText(str(srt), left + 8, top)
                rect = (rect[0], top + 2, rect[2], rect[3])
                s = self.GetColLabelValue(col)
                if s.find("\n") >= 0:
                    font.SetPointSize(fs2)
                else:
                    font.SetPointSize(fs)
                dc.SetFont(font)
                dc.DrawLabel(
                    "%s" % self.GetColLabelValue(col), rect, alignment=wx.ALIGN_CENTER
                )
            else:
                font.SetWeight(wx.NORMAL)
                font.SetPointSize(fs)
                dc.SetFont(font)
                dc.DrawLabel(
                    "%s" % self.GetColLabelValue(col), rect, alignment=wx.ALIGN_CENTER
                )

    def on_select_cell(self, evt):
        newrow = evt.GetRow()
        if self.oldrow != newrow:
            if self.panel:
                self.panel.refresh(newrow)
            self.SelectRow(evt.GetRow())
        evt.Skip()

    def on_range_selected(self, evt):
        if evt.Selecting():
            if evt.GetTopRow() != evt.GetBottomRow():
                self.ClearSelection()
                for row in range(evt.GetTopRow(), evt.GetBottomRow() + 1):
                    self.GetTable().sel_row(row)
                if self.GetGridCursorRow() == evt.GetTopRow():
                    self.SelectRow(evt.GetBottomRow())
                    self.SetGridCursor(evt.GetBottomRow(), self.GetGridCursorCol())
                else:
                    self.SelectRow(evt.GetTopRow())
                    self.SetGridCursor(evt.GetTopRow(), self.GetGridCursorCol())
        evt.Skip()

    def on_cell_left_click(self, evt):
        row1 = self.GetGridCursorRow()
        row2 = evt.GetRow()
        if evt.ControlDown():
            self.GetTable().sel_row(row2)
        elif evt.ShiftDown():
            if row1 > row2:
                (row1, row2) = (row2, row1)
            for row in range(row1, row2 + 1):
                self.GetTable().sel_row(row)
        # elif not evt.ControlDown() and not evt.ShiftDown():
        else:
            s = self.GetTable().GetValue(evt.GetRow(), evt.GetCol())
            if type(s) == str and s.strip() == "+":
                attr = self.GetTable().get_ext_attr(evt.GetRow(), evt.GetCol())
                td = self.GetTable().data[evt.GetRow()][evt.GetCol()]
                children = self.GetTable().get_children(evt.GetRow(), evt.GetCol())
                if children:
                    for child_id in children:
                        child = children[child_id]
                        if "href" in child.attrs:
                            self.GetParent().GetParent().href_clicked(self, child.attrs)
                            break
            else:
                if self.typ == self.GET_ID:
                    self.action("get")
                else:
                    self.SelectRow(evt.GetRow())
                    self.SetGridCursor(evt.GetRow(), evt.GetCol())
                    self.SetFocus()
                #    evt.Skip()

    def on_l_up(self, evt):
        evt.Skip()

    def on_cell_left_dclick(self, evt):
        if self.typ == self.VIEW or self.typ == self.GET_ID or self.readonly:
            self.SelectRow(evt.GetRow())
            self.SetGridCursor(evt.GetRow(), evt.GetCol())
            self.SetFocus()
            value = self.GetTable().GetValue(self.GetGridCursorRow(), 0)
            if value and value.strip() == "-":
                self.action("insert")
            else:
                if self.typ == self.GET_ID:
                    self.action("get")
                else:
                    if self.action_exists("get_row"):
                        self.action("get_row")
                    else:
                        self.action(self.default_command)
        else:
            self.SetGridCursor(evt.GetRow(), evt.GetCol())
            self.begin_edit()

    def on_key_j(self, evt):
        row = self.GetGridCursorRow()
        if row + 1 < self.GetTable().GetNumberRows():
            self.MoveCursorDown(False)

    def on_key_k(self, evt):
        row = self.GetGridCursorRow()
        if row > 0:
            self.MoveCursorUp(False)

    def on_key_h(self, evt):
        col = self.GetGridCursorCol()
        if col > 0:
            self.MoveCursorLeft(False)

    def on_key_l(self, evt):
        col = self.GetGridCursorCol()
        if col + 1 < self.GetTable().GetNumberCols():
            self.MoveCursorRight(False)

    def on_key_down(self, evt):
        if (
            evt.KeyCode in (wx.WXK_UP, wx.WXK_DOWN, wx.WXK_LEFT, wx.WXK_RIGHT)
            and not evt.AltDown()
            and not evt.ControlDown()
            and not evt.ShiftDown()
        ):
            row = self.GetGridCursorRow()
            col = self.GetGridCursorCol()
            if evt.KeyCode == wx.WXK_DOWN:
                self.on_key_j(evt)
            if evt.KeyCode == wx.WXK_UP:
                self.on_key_k(evt)
            if evt.KeyCode == wx.WXK_RIGHT:
                self.on_key_l(evt)
            if evt.KeyCode == wx.WXK_LEFT:
                self.on_key_h(evt)
            return

        if evt.AltDown() and evt.KeyCode != wx.WXK_RETURN:
            wx.GetApp().GetTopWindow().on_acc_key_down(evt)
            return
        if evt.KeyCode == wx.WXK_TAB and evt.ControlDown():
            evt.Skip()
            return
        if evt.KeyCode == wx.WXK_RETURN and not evt.AltDown():
            if self.typ == self.VIEW:
                value = self.GetTable().GetValue(self.GetGridCursorRow(), 0)
                if value and value.strip() == "-":
                    self.action("insert")
                else:
                    if self.action_exists("get_row"):
                        self.action("get_row")
                    else:
                        self.action(self.default_command)
                return
            if self.typ == self.GET_ID:
                if hasattr(self.Parent, "ReturnRec"):
                    rec = self.GetTable().get_rec(self.GetGridCursorRow())
                    nr = rec[0]
                    s = self.GetTable().GetRecAsStr(nr)
                    self.Parent.ReturnRec((nr, s, rec))
                else:
                    self.action("get")
                return
            if evt.ControlDown():
                evt.Skip()
                return
            self.DisableCellEditControl()
            success = self.MoveCursorRight(evt.ShiftDown())
            if not success:
                new_row = self.GetGridCursorRow() + 1
                if new_row < self.GetTable().GetNumberRows():
                    self.SetGridCursor(new_row, 0)
                    self.MakeCellVisible(new_row, 0)
                else:
                    pass
            return
        if evt.KeyCode == wx.WXK_DELETE:
            if self.typ == self.VIEW or self.typ == self.GET_ID:
                if not self.readonly:
                    self.action("delete")
                evt.Skip()
                return
            if not evt.ControlDown():
                evt.Skip()
                return
            row = self.GetGridCursorRow()
            if row >= 0:
                self.DeleteRows(row, 1)
                if row < self.GetTable().get_rec_count():
                    self.SetGridCursor(row + 1, 0)
                    self.MakeCellVisible(row + 1, 0)
        if evt.KeyCode == wx.WXK_F4 or evt.KeyCode == wx.WXK_RETURN and evt.AltDown():
            row = self.GetGridCursorRow()
            rec = self.GetTable().get_rec(row)
            names = self.GetTable().GetColNames()[1:]
            parm = dict()
            parm["rec"] = rec
            self.GetHtmlParent().ActiveCtrl = self
            okno = self.GetHtmlParent().new_child_page(
                self.address + "../" + str(rec[0]) + "/edit/", "", param=parm
            )
            return
        if evt.KeyCode == wx.WXK_HOME and not evt.ControlDown():
            self.goto_first_row()
            return
        if evt.KeyCode == wx.WXK_END and not evt.ControlDown():
            self.goto_last_row()
            return
        if evt.KeyCode == wx.WXK_SPACE:
            row = self.GetGridCursorRow()
            if row >= 0:
                self.GetTable().sel_row(row)
                if row < self.GetNumberRows() - 1:
                    self.SetGridCursor(row + 1, self.GetGridCursorCol())
                return
        if (evt.KeyCode == ord("S") or evt.KeyCode == ord("s")) and evt.ControlDown():
            self.GetTable().commit()
            self.GetTable().refresh(True)
            evt.Skip()
            return
        if (evt.KeyCode == ord("R") or evt.KeyCode == ord("r")) and evt.ControlDown():
            self.GetTable().refresh(True)
            evt.Skip()
            return

        if evt.ControlDown() and evt.GetKeyCode() in [67, 322]:
            self.copy(evt)
            return

            # Ctrl+V
        elif evt.ControlDown() and evt.GetKeyCode() == 86:
            self.paste(evt)
            return

        evt.Skip()
        return

    def copy(self, evt, format="json"):
        x = self.GetTable().copy()
        if x:
            if wx.TheClipboard.Open():
                if format == "csv":
                    data = []
                    first_rec = True
                    for obj in x:
                        row = obj["fields"]
                        if first_rec:
                            rec = []
                            for key, value in row.items():
                                rec.append(key)
                            data.append(rec)
                            first_rec = False
                        rec = []
                        for pos in data[0]:
                            if pos in row:
                                rec.append(row[pos])
                            else:
                                rec.append(None)
                        data.append(rec)
                    writer = io.StringIO()
                    csvwriter = csv.writer(writer)
                    for row in data:
                        csvwriter.writerow(row)
                    writer.seek(0)
                    x = writer.read()
                    wx.TheClipboard.SetData(wx.TextDataObject(x))
                if format == "json":
                    wx.TheClipboard.SetData(wx.TextDataObject(json_dumps(x, indent=4)))
                wx.TheClipboard.Close()
            else:
                wx.MessageBox(_("Unable to open the clipboard"), _("Error"))

    def paste(self, evt):
        text_data = wx.TextDataObject()
        if wx.TheClipboard.Open():
            success = wx.TheClipboard.GetData(text_data)
            wx.TheClipboard.Close()
            if success:
                txt = text_data.GetText()
                ret = self.GetTable().paste(json_loads(txt))
                if "success" in ret:
                    self.GetTable().refresh(True)
                    self.ForceRefresh()
        else:
            wx.MessageBox(_("Unable to open the clipboard"), _("Error"))

    def on_update_rec_from_form(self, rec):
        row = self.GetGridCursorRow()
        names = self.GetTable().GetColNames()[1:]
        for i in range(len(names)):
            name = names[i]
            self.GetTable().SetValue(row, i, rec[name])
        self.ForceRefresh()

    def accepts_focus(self):
        return True

    def refr_count(self, old_count, count, store_pos=0, autosize=True):
        dy = old_count - count
        if dy != 0:
            if dy > 0:
                msg = GridTableMessage(
                    self.GetTable(), GRIDTABLE_NOTIFY_ROWS_DELETED, count, dy
                )
                self.ProcessTableMessage(msg)
            else:
                if old_count == 0 and not autosize:
                    dyy = -1 * dy
                    if dyy > 128:
                        dyy = 128
                    msg = GridTableMessage(
                        self.GetTable(), GRIDTABLE_NOTIFY_ROWS_APPENDED, dyy
                    )
                    self.ProcessTableMessage(msg)
                    # self.AutoSizeColumns(True)
                    dyy = -1 * dy
                    if dyy > 128:
                        msg = GridTableMessage(
                            self.GetTable(), GRIDTABLE_NOTIFY_ROWS_APPENDED, dyy - 128
                        )
                        self.ProcessTableMessage(msg)
                else:
                    dyy = -1 * dy
                    msg = GridTableMessage(
                        self.GetTable(), GRIDTABLE_NOTIFY_ROWS_APPENDED, dyy
                    )
                    self.ProcessTableMessage(msg)

            if count > 0:
                if store_pos == 0:
                    self.SetGridCursor(0, 0)
                    self.MakeCellVisible(0, 0)
                elif store_pos == 1:
                    row = self.GetGridCursorRow()
                    if row >= count:
                        self.SetGridCursor(0, 0)
                        self.MakeCellVisible(0, 0)
                else:
                    row = self.GetGridCursorRow()
                    if row >= count:
                        self.SetGridCursor(count - 1 + self.CanAppend, 0)
                        self.MakeCellVisible(count - 1 + self.CanAppend, 0)

    def on_label_left_click(self, evt):
        if evt.GetCol() >= 0:
            str = self.GetColLabelValue(evt.GetCol())
            if evt.ControlDown():
                self.GetTable().sort(evt.GetCol(), True)
            else:
                self.GetTable().sort(evt.GetCol(), False)
            evt.Skip()

    def on_label_right_click(self, evt):
        message = wx.MessageDialog(self, "x", "Komunikat")
        message.ShowModal()

    def on_label_left_dclick(self, evt):
        message = wx.MessageDialog(self, "y", "Komunikat")
        message.ShowModal()

    def on_cell_right_click(self, evt):
        message = wx.MessageDialog(self, "a", "Komunikat")
        message.ShowModal()

    def action_exists(self, command):
        action = self.GetTable().get_actions(self.GetGridCursorRow())
        if action and command in action:
            return True
        else:
            return False

    def action(self, command):
        action = self.GetTable().get_actions(self.GetGridCursorRow())
        if action and command in action:
            if action[command][0] == "openurl":
                self.last_action = command
                if command == "get_row":
                    p = action[command]
                    id = p[3]["data-id"]
                    title = p[3]["data-text"]
                    page = self.GetParent().get_parent_page().get_parent_page()
                    if page:
                        page.signal("return_updated_row", id=id, title=title)
                        wx.CallAfter(self.GetParent().get_parent_form().cancel)
                    return 1
                else:
                    ret = (
                        self.GetParent()
                        .GetParent()
                        .href_clicked(self, action[command][3])
                    )
                    ret = 1
                    return ret

            if action[command][0] == "tableurl":
                self.GetTable().proxy.set_address(action[command][1])
                if hasattr(self.GetParent(), "table_url"):
                    self.GetParent().table_url(action[command][1])
                self.GetTable().refresh(True)

            if action[command][0] == "command":
                if hasattr(self.GetParent(), "table_command"):
                    self.GetParent().table_command(action[command][1])

    def get_action_list(self, row=0):
        akcje = self.GetTable().get_action_list(row)
        if akcje:
            return (not self.readonly, akcje)
        else:
            return (not self.readonly, None)

    def get_actions(self, row=0):
        action = self.GetTable().get_actions(row)
        if action:
            return (not self.readonly, list(action.items()))
        else:
            return (not self.readonly, None)

    def get_table_and_state(self):
        tableandstate = []
        tableandstate.append((self.GetGridCursorRow(), self.GetGridCursorCol()))
        tableandstate.append(self.GetTable().get_table_and_state())
        return tableandstate

    def duplicate_table_and_state(self):
        tableandstate = []
        tableandstate.append((0, 0))
        tableandstate.append(self.GetTable().duplicate_table_and_state())
        return tableandstate

    def set_table_and_state(self, tableandstate):
        oldrowcount = self.GetTable().last_row_count
        oldtableandstate = self.get_table_and_state()
        self.SetGridCursor(tableandstate[0][0], tableandstate[0][1])
        self.GetTable().set_table_and_state(tableandstate[1])
        rowcount = self.GetTable().GetNumberRows()
        self.refr_count(oldrowcount, rowcount, store_pos=1)
        self.ForceRefresh()
        return oldtableandstate



"""
SchNotebook object used as a top window in panels: 'desktop', 'panel', 'header' and 'footer'
"""

import wx
import wx.lib.agw.aui as aui
from wx.lib.agw.aui import framemanager
from pytigon_lib.schtools.wiki import wiki_from_str
from pytigon_gui.guiframe.manager import SChAuiBaseManager
from wx.lib.agw.aui.aui_constants import *


class SchNotebook(aui.AuiNotebook):
    """SchNotebook class"""

    def __init__(
        self,
        parent,
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.BORDER_NONE | wx.TAB_TRAVERSAL | wx.WANTS_CHARS,
    ):
        """Constructor

        Args:
            parent: parent window
            pos: position of window
            size: size of window
            style: see wx.Window style
        """
        self._curpage = -1
        self._tab_id_counter = AuiBaseTabCtrlId
        self._dummy_wnd = None
        self._hide_tabs = False
        self._sash_dclick_unsplit = False
        self._tab_ctrl_height = 20
        self._requested_bmp_size = wx.Size(-1, -1)
        self._requested_tabctrl_height = -1
        self._textCtrl = None
        self._tabBounds = (-1, -1)

        self.panel = None
        self.active = False
        self.closing = False
        self.last_active = None

        wx.Panel.__init__(self, parent, wx.ID_ANY, pos, size, style, name="SchNotebook")
        self._mgr = SChAuiBaseManager()

        self._tabs = aui.AuiTabContainer(self)
        self.InitNotebook(AUI_NB_DEFAULT_STYLE)

        self.Bind(aui.EVT_AUINOTEBOOK_PAGE_CLOSE, self.on_closing)
        self.Bind(aui.EVT_AUINOTEBOOK_TAB_DCLICK, self.on_dclick)
        self.Bind(aui.EVT_AUINOTEBOOK_PAGE_CHANGED, self.on_changed)
        self.Bind(aui.EVT_AUINOTEBOOK_PAGE_CHANGING, self.on_changing)
        self.Bind(wx.EVT_NAVIGATION_KEY, self.on_navigete)

        self.SetWindowStyleFlag(wx.WANTS_CHARS)
        self.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND))

    def on_closing(self, event):
        idn = event.GetSelection()
        if idn >= 0:
            page = self.GetPage(idn)
            self.closing = True
            if page.close_no_del(True):
                self.GetParent().GetParent().after_close(self)
                event.Veto()

    def on_changed(self, event):
        sel = event.GetSelection()
        old_sel = event.GetOldSelection()
        if old_sel >= 0:
            self.GetPage(old_sel).Refresh()
        if sel >= 0:
            self.GetPage(sel).Refresh()
            self.activate_page(self.GetPage(sel))
        event.Skip()

    def on_changing(self, event):
        if self.closing:
            self.closing = False
            sel = event.GetSelection()
            oldsel = event.GetOldSelection()
            if oldsel == -1:
                new_page = self.GetPage(sel)
                tab = self.FindTab(new_page)[0]
                if tab:
                    count = tab.GetPageCount()
                    if count > 0:
                        new_sel = count - 1
                        if new_sel > sel:
                            self.SetSelection(new_sel)
                            event.Veto()
                            return
        event.Skip()

    def DeletePage(self, sel):
        page = self.GetPage(sel)
        if page == self.last_active:
            self.last_active = None

        def _close():
            self.panel.Hide()
            wx.GetApp().GetTopWindow()._mgr.Update()
            for pane_name in ["desktop", "panel", "menu", "header", "footer"]:
                pane_info = wx.GetApp().GetTopWindow()._mgr.GetPane(pane_name)
                if pane_info.IsOk() and pane_info.IsShown():
                    pane_info.window.SetFocus()

        if len(self._tabs._pages) == 1:
            wx.CallAfter(_close)

        return aui.AuiNotebook.DeletePage(self, sel)

    def on_dclick(self, event):
        try:
            txt = self.GetPageText(event.GetSelection())
            mp, adr = wx.GetApp().read_html(
                self, "/schwiki/help/" + wiki_from_str(txt) + "/view/", None
            )
            if not txt.startswith("?"):
                wiki = wiki_from_str(txt)
                wx.GetApp().GetTopWindow().new_main_page(
                    mp, "?: " + wiki, panel="desktop2"
                )
        except:
            pass

    def SetPanel(self, panel):
        self.panel = panel

    def GetPanel(self):
        return self.panel

    def set_active(self, active):
        self.active = active
        if active:
            w = self.GetCurrentPage()
            self.activate_page(w)
            self.GetParent().GetParent().SetActive(self, w)

    def activate_page(self, page):
        if self.last_active:
            self.last_active.deactivate_page()
        self.last_active = page
        if page:
            page.activate_page()

    def add_and_split(self, page, title, direction):
        """Add page to the notebook and split notebook window

        Args:
            page - page to be added
            title - title of new tab
            direction - split direction: wx.LEFT, wx.TOP, wx.RIGHT, wx.Bottom
        """
        self.closing = False
        if self.GetPageCount() < 1:
            return
        cli_size = self.GetClientSize()
        if self.GetPageCount() > 2:
            split_size = self.CalculateNewSplitSize()
        else:
            split_size = self.GetClientSize()
            split_size.x = int(split_size.x / 2)
            split_size.y = int(split_size.y / 2)
        new_tabs = aui.TabFrame(self)
        new_tabs.SetTabCtrlHeight(self._tab_ctrl_height)
        self._tab_id_counter += 1
        new_tabs._tabs = aui.AuiTabCtrl(self, self._tab_id_counter)
        new_tabs._tabs.SetArtProvider(self._tabs.GetArtProvider().Clone())
        new_tabs._tabs.SetAGWFlags(self._agwFlags)
        dest_tabs = new_tabs._tabs
        pane_info = (
            framemanager.AuiPaneInfo()
            .CaptionVisible(False)
            .BestSize(split_size.GetWidth(), split_size.GetHeight())
        )
        if direction == wx.LEFT:
            pane_info.Left()
            mouse_pt = wx.Point(0, int(cli_size.y / 2))
        elif direction == wx.RIGHT:
            pane_info.Right()
            mouse_pt = wx.Point(cli_size.x, int(cli_size.y / 2))
        elif direction == wx.TOP:
            pane_info.Top()
            mouse_pt = wx.Point(int(cli_size.x / 2), 0)
        elif direction == wx.BOTTOM:
            pane_info.Bottom()
            mouse_pt = wx.Point(int(cli_size.x / 2), cli_size.y)
        # self._mgr.AddPane(new_tabs, pane_info, mouse_pt)
        # self._mgr.Update()
        page_info = aui.AuiNotebookPage()
        page_info.window = page
        page_info.caption = title
        page_info.active = False
        page_info.control = None
        idn = self.GetPageCount()
        self._tabs.InsertPage(page, page_info, idn)
        dest_tabs.AddPage(page, page_info)
        self.SetPageTextColour(idn, wx.Colour(0, 96, 0))
        page.Reparent(self)
        self.DoSizing()
        dest_tabs.Refresh()
        # self.SetSelectionToPage(page_info)
        self.UpdateHintWindowSize()

        self._mgr.AddPane(new_tabs, pane_info, mouse_pt)
        self._mgr.Update()

    def on_navigete(self, evt):
        forward = evt.GetDirection()
        self.activate_page(self.GetCurrentPage())
        evt.Skip()

    def Freeze(self):
        pass

    def Thaw(self):
        pass



import wx
from wx.lib.agw import customtreectrl as CT
from wx.lib.agw.ribbon import art

from pytigon_gui.guilib.events import *
from pytigon_gui.toolbar.basetoolbar import (
    BaseHtmlPanel,
    ToolbarBar,
    ToolbarPage,
    ToolbarPanel,
    ToolbarButton,
)
from pytigon_gui.guictrl.basectrl import SchBaseCtrl


_ = wx.GetTranslation


def _tree_list(tree, parent_item, list=None):
    if list == None:
        list2 = []
    else:
        list2 = list

    item, cookie = tree.GetFirstChild(parent_item)
    while item and item.IsOk():
        list2.append(item)
        if tree.ItemHasChildren(item):
            _tree_list(tree, item, list2)
        item, cookie = tree.GetNextChild(parent_item, cookie)

    return list2


class TreeHtmlPanel(BaseHtmlPanel):
    def get_width(self):
        return self.page.parent_bar.get_bar_width()

    def get_height(self):
        return self.page.parent_bar.get_bar_height()

    def set_page(self, html_page):
        super().set_page(html_page)
        # self.page.parent_bar.update()
        # self.page.parent_bar.SetActivePage(self.page)


class TreeToolbarButton(ToolbarButton):
    def __init__(
        self,
        parent_panel,
        id,
        title,
        bitmap,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        ToolbarButton.__init__(
            self, parent_panel, id, title, bitmap, bitmap_disabled, kind
        )


class TreeToolbarPanel(ToolbarPanel):
    def __init__(self, parent_page, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        ToolbarPanel.__init__(self, parent_page, title, kind)
        bar = self.parent_page.parent_bar
        self.tree_item = bar.append_to_tree(parent_page.tree_item, title)

    def _append(self, b):
        item = None
        if self.kind in (
            ToolbarPanel.TYPE_PANEL_TOOLBAR,
            ToolbarPanel.TYPE_PANEL_BUTTONBAR,
        ):
            if b.kind == ToolbarButton.TYPE_SIMPLE:
                item = self.parent_page.parent_bar.append_to_tree(
                    self.tree_item, b.title, b.bitmap, ct_type=0
                )
            elif b.kind == ToolbarButton.TYPE_DROPDOWN:
                item = self.parent_page.parent_bar.append_to_tree(
                    self.tree_item, b.title, b.bitmap, ct_type=0
                )
            elif b.kind == ToolbarButton.TYPE_HYBRID:
                item = self.parent_page.parent_bar.append_to_tree(
                    self.tree_item, b.title, b.bitmap, ct_type=0
                )
            elif b.kind == ToolbarButton.TYPE_TOOGLE:
                item = self.parent_page.parent_bar.append_to_tree(
                    self.tree_item, b.title, b.bitmap, ct_type=1
                )
            elif b.kind == ToolbarButton.TYPE_PANEL:
                pass
            elif b.kind == ToolbarButton.TYPE_SEPARATOR:
                item = self.parent_page.parent_bar.AppendSeparator(self.tree_item)

        if item:
            self.parent_page.parent_bar.SetItemHyperText(item, True)
            self.parent_page.parent_bar.SetPyData(item, b.id)

    def create_button(
        self,
        id,
        title,
        bitmap=None,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        b = TreeToolbarButton(self, id, title, bitmap, bitmap_disabled, kind)
        self._append(b)
        return b

    def add_separator(self):
        b = TreeToolbarButton(
            self, 0, "", None, None, kind=ToolbarButton.TYPE_SEPARATOR
        )
        self._append(b)
        return b


class TreeToolbarPage(ToolbarPage):
    def __init__(self, parent_bar, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        ToolbarPage.__init__(self, parent_bar, title, kind)
        self.tree_item = parent_bar.append_to_tree(parent_bar.tree_main_page, title)

    def create_panel(self, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        return TreeToolbarPanel(self, title, kind)

    def create_html_panel(self, title):
        p = wx.Panel(self.parent_bar, size=wx.Size(200, 200))
        self.parent_bar.AppendItem(self.tree_item, title, 0, p)
        self.parent_bar.Expand(self.tree_item)
        return TreeHtmlPanel(self, p)


class TreeToolbarBar(ToolbarBar, CT.CustomTreeCtrl):
    def __init__(self, parent, gui_style):
        agwStyle = (
            CT.TR_HIDE_ROOT
            | CT.TR_HAS_BUTTONS
            | CT.TR_HAS_VARIABLE_ROW_HEIGHT
            | CT.TR_NO_LINES
        )
        CT.CustomTreeCtrl.__init__(self, parent, agwStyle=agwStyle)

        self._indent = 8
        self.tree_main_page = self.AddRoot("The Root Item")
        self.images = wx.ImageList(32, 32)
        self.image_id = 0
        self.show_titles = True

        c2 = wx.SystemSettings.GetColour(wx.SYS_COLOUR_MENU)
        m = 1.0
        x1 = int((int(c2.Red() * m) + 255) / 2)
        x2 = int((int(c2.Green() * m) + 255) / 2)
        x3 = int((int(c2.Blue() * m) + 255) / 2)
        c = wx.Colour(x1, x2, x3)
        self.bg = c
        self.SetBackgroundColour(c)
        c3 = wx.Colour(0, 0, 0)
        self._hypertextnewcolour = c3
        self._hypertextvisitedcolour = c3

        ToolbarBar.__init__(self, parent, gui_style)
        self.EnableSelectionVista(True)

        self.Bind(wx.EVT_CLOSE, self.on_close)
        self.Bind(CT.EVT_TREE_ITEM_HYPERLINK, self.on_hyper_link)

        wx.GetApp().GetTopWindow().idle_objects.append(self)

    def append_to_tree(self, parent_elem, title, bitmap=None, ct_type=0):
        if bitmap is not None and bitmap.IsOk():
            if bitmap.GetWidth() < 32 or bitmap.GetHeight() < 32:
                b = wx.BitmapFromImage(bitmap.ConvertToImage().Rescale(32, 32))
            else:
                b = bitmap
            self.images.Add(b)
            item = self.AppendItem(
                parent_elem,
                title if self.show_titles else "",
                image=self.image_id,
                ct_type=ct_type,
            )
            self.image_id += 1
        else:
            item = self.AppendItem(
                parent_elem, title if self.show_titles else "", ct_type=ct_type
            )
        return item

    def create_page(self, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        return TreeToolbarPage(self, title, kind)

    def CanAcceptFocus(self):
        return True

    def CanAcceptFocusFromKeyboard(self):
        return self.CanAcceptFocus()

    def on_close(self, event):
        if self in wx.GetApp().GetTopWindow().idle_objects:
            del wx.GetApp().GetTopWindow().idle_objects[self]
        event.Skip()

    def get_max_width(self, respect_expansion_state=True):
        return 400

    def realize(self):
        self.AssignImageList(self.images)
        self.ExpandAll()
        self.SetSize((300, 400))
        x = self.GetMaxWidth()

    def on_hyper_link(self, event):
        wx.Event.EventType
        item = event.GetItem()
        if item:
            wx.CallAfter(self.send_menu_event, item)

    def send_menu_event(self, item):
        e = wx.CommandEvent(wx.EVT_MENU.typeId, self.GetPyData(item))
        ret = self.ProcessEvent(e)
        if not ret:
            win = wx.GetApp().GetTopWindow().get_active_ctrl()
            if win and issubclass(type(win), SchBaseCtrl):
                win.ProcessEvent(e)

    def on_idle(self):
        refresh = False
        for item in _tree_list(self, self.tree_main_page):
            id = self.GetPyData(item)
            if id:
                if (id >= ID_START and id < ID_END) or (
                    id >= wx.ID_LOWEST and id < wx.ID_HIGHEST
                ):
                    event = wx.UpdateUIEvent(id)
                    event.Enable(False)
                    event.SetEventObject(self.GetParent())

                    if self.ProcessWindowEvent(event):
                        if item.IsEnabled() != event.GetEnabled():
                            refresh = True
                        item.Enable(event.GetEnabled())
                    else:
                        enable = False
                        win = wx.Window.FindFocus()
                        if win and issubclass(type(win), SchBaseCtrl):
                            if win.ProcessEvent(event):
                                if event.GetEnabled():
                                    enable = True
                        if enable:
                            if not item.IsEnabled():
                                refresh = True
                            item.Enable(True)
                        else:
                            if item.IsEnabled():
                                refresh = True
                            item.Enable(False)
        if refresh:
            self.Refresh()

    def set_active_page(self, page):
        pass

    def get_bar_height(self):
        return 600

    def get_bar_width(self):
        return 150

    def remove_page(self, title):
        for p in self.pages:
            if p.title == title:
                self.Delete(p.tree_item)
                break
        super().remove_page(title)

    def create(self):
        self.AssignImageList(self.images)
        self.ExpandAll()
        self.SetSize((300, 400))
        x = self.GetMaxWidth()

    def close(self):
        self.Close()

    def update(self):
        size = self.GetSize()
        self.SetSize(wx.Size(size.GetWidth() - 1, size.GetHeight()))
        self.SetSize(wx.Size(size.GetWidth(), size.GetHeight()))

    def bind_ui(self, fun, id=wx.ID_ANY):
        self.parent.Bind(wx.EVT_UPDATE_UI, fun, id=id)

    def bind(self, fun, id=wx.ID_ANY, e=None):
        if e:
            self.parent.Bind(e, fun, id=id)
        else:
            self.parent.Bind(wx.EVT_MENU, fun, id=id)

    def bind_dropdown(self, fun, id):
        self.parent.Bind(wx.EVT_MENU, fun, id=id)

    def un_bind(self, id, e=None):
        if e:
            self.Unbind(e, id=id)
        else:
            self.Unbind(wx.EVT_MENU, id=id)



import wx
import sys

import pytigon_gui.guictrl.ctrl

_ = wx.GetTranslation


class HttpErrorDialog(wx.Dialog):
    def __init__(
        self,
        parent,
        title,
        text,
        size=wx.DefaultSize,
        pos=wx.DefaultPosition,
        style=wx.DEFAULT_DIALOG_STYLE,
        use_metal=False,
    ):
        try:
            pre = wx.PreDialog()
            pre.SetExtraStyle(wx.DIALOG_EX_CONTEXTHELP)
            pre.Create(parent, wx.ID_ANY, title, pos, size, style)
            self.PostCreate(pre)
        except:
            wx.Dialog.__init__(self, parent, wx.ID_ANY, title, pos, size, style)

        if "wxMac" in wx.PlatformInfo and use_metal:
            self.SetExtraStyle(wx.DIALOG_EX_METAL)
        sizer = wx.BoxSizer(wx.VERTICAL)

        self.label = pytigon_gui.guictrl.ctrl.HTML2(
            self, size=(1024, 768), name="webbrowser"
        )

        try:
            sizer.Add(self.label.wb, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        except:
            sizer.Add(self.label, 0, wx.ALIGN_CENTRE | wx.ALL, 5)

        btnsizer = wx.StdDialogButtonSizer()
        btn = wx.Button(self, wx.ID_OK, _("Continue"))
        btn.SetDefault()
        btnsizer.AddButton(btn)
        btn = wx.Button(self, wx.ID_CANCEL, _("Break"))
        btn.SetHelpText(_("The Break button breaks the application"))
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        # sizer.Add(btnsizer, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)
        sizer.Add(btnsizer, 0, wx.ALL, 5)
        self.SetSizer(sizer)
        sizer.Fit(self)
        wx.CallAfter(self.label.load_str, text)
        # self.label.load_str(text)

    def set_acc_key_tab(self, ctrl, tab):
        pass


def _http_error(parent, content):
    """Show form with error content returned by http server

    Args:
        parent - parent window
        content - html page returned by http server
    """
    lock = wx.GetApp().lock
    if not lock:
        wx.GetApp().lock = True

        if parent and hasattr(parent, "Invalidate"):
            parent.Invalidate()

        if type(content) == str:
            c = content
        else:
            c = content.decode("utf-8")
        dlg = HttpErrorDialog(
            wx.GetApp().GetTopWindow(), _("Error message"), c, size=(800, 600)
        )
        dlg.CenterOnScreen()
        try:
            val = dlg.ShowModal()
        except:
            print("ERROR")
        wx.GetApp().lock = False

        if val == wx.ID_CANCEL:
            sys.exit()


def http_error(parent, content):
    wx.CallAfter(_http_error, parent, content)



import wx
import wx.html
from pytigon_lib.schtools import schjson


schEVT_THREAD_INFO = wx.NewEventType()
EVT_THREAD_INFO = wx.PyEventBinder(schEVT_THREAD_INFO, 1)


class ThreadEvent(wx.PyCommandEvent):
    def __init__(self, evt_type, id):
        wx.PyCommandEvent.__init__(self, evt_type, id)
        self.info = None

    def set_info(self, val):
        self.info = val

    def get_info(self):
        return self.info


class SchThreadWindow(wx.Panel):
    def __init__(self, manager, thread_name, *args, **kwds):
        wx.Panel.__init__(self, *args, **kwds)
        self.thread_name = thread_name
        self.manager = manager
        self.closed = False
        self.SetBackgroundColour(wx.NamedColour("#eec"))
        self.gauge = wx.Gauge(self, range=100, size=(100, 1))
        self.gauge.SetValue(0)
        self.html = wx.html.HtmlWindow(
            self, size=(50, 50), style=wx.html.HW_SCROLLBAR_NEVER
        )
        self.html.SetBorders(0)
        self.html.SetPage("<body bgcolor='#eec'></body>")
        bmp = wx.ArtProvider.GetBitmap(wx.ART_GO_TO_PARENT, wx.ART_TOOLBAR, (16, 16))
        bmp = bmp.ConvertToImage().Rescale(16, 8).ConvertToBitmap()
        bmp2 = wx.ArtProvider.GetBitmap(wx.ART_QUIT, wx.ART_TOOLBAR, (16, 16))
        bmp2 = bmp2.ConvertToImage().Rescale(16, 8).ConvertToBitmap()
        button1 = wx.BitmapButton(self, -1, bmp)
        button2 = wx.BitmapButton(self, -1, bmp2)
        self.Bind(wx.EVT_BUTTON, self.on_expand, button1)
        self.Bind(wx.EVT_BUTTON, self.on_kill, button2)
        box = wx.BoxSizer(wx.HORIZONTAL)
        box.Add(self.html, 1, wx.EXPAND)
        box.Add(self.gauge, 0, wx.EXPAND)
        box.Add(button1, 0, wx.EXPAND)
        box.Add(button2, 0, wx.EXPAND)
        self.SetSizer(box)
        box.Fit(self)

    def is_closed(self):
        return self.closed

    def timer(self):
        http = wx.GetApp().http
        response = http.get(
            self, "http://local.net/schsys/thread_short_info/" + self.thread_name
        )
        info_json = response.str()
        info = schjson.loads(info_json)

        if type(info) == str and info == "$$$":
            self.closed = True
            evt = ThreadEvent(schEVT_THREAD_INFO, -1)
            evt.set_info(self.thread_name)
            wx.GetApp().GetTopWindow().GetEventHandler().ProcessEvent(evt)
        else:
            if info:
                if "progress" in info:
                    progress = info["progress"]
                    self.gauge.SetValue(int(progress))
                if "description" in info:
                    description = info["description"]
                    self.html.SetPage("<body bgcolor='#eec'>" + description + "</body>")

    def on_expand(self, event):
        address = "http://local.net/schsys/thread_long_info/" + self.thread_name
        wx.GetApp().GetTopWindow().new_main_page(address, "aplikacja")

    def on_kill(self, event):
        http = wx.GetApp().http
        http.get(self, "http://local.net/schsys/thread_kill/" + self.thread_name)


class SchThreadManager(object):
    def __init__(self, app, statusbar):
        self.app = app
        self.statusbar = statusbar
        self.sizeChanged = True
        self.windows = []
        statusbar.Bind(wx.EVT_SIZE, self.on_size)
        statusbar.Bind(wx.EVT_IDLE, self.on_idle)

    def append(self, thread_address):
        self.windows.append(SchThreadWindow(self, thread_address, self.statusbar))

    def reposition(self):
        x = len(self.windows)
        if x > 0:
            rect = self.statusbar.GetFieldRect(1)
            w = 0
            for win in self.windows:
                win.SetPosition((rect.x + w + 2, rect.y + 2))
                win.SetSize((rect.width / x - 4, rect.height - 3))
                w += rect.width / x

    def on_size(self, evt):
        self.reposition()
        self.sizeChanged = True

    def on_idle(self, evt):
        if self.sizeChanged:
            self.reposition()

    def timer(self):
        if len(self.windows) > 0:
            for win in self.windows:
                if win.is_closed():
                    win.Destroy()
                    self.windows.remove(win)
                else:
                    win.timer()



import wx

from pytigon_gui.guilib.events import *
from pytigon_gui.toolbar.basetoolbar import (
    ToolbarBar,
    ToolbarPage,
    ToolbarPanel,
    ToolbarButton,
)


class MenuToolbarButton(ToolbarButton, wx.MenuItem):
    def __init__(
        self,
        parent_panel,
        id,
        title,
        bitmap=None,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        wx.MenuItem.__init__(self, parentMenu=parent_panel, id=id, text=title)
        if bitmap and bitmap.IsOk():
            self.SetBitmap(bitmap)


class MenuToolbarPanel(ToolbarPanel, wx.Menu):
    def __init__(self, parent_page, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        wx.Menu.__init__(self)
        ToolbarPanel.__init__(self, parent_page, title, kind)
        self.parent_page.parent_bar.Append(self, title)

    def create_button(
        self,
        id,
        title,
        bitmap=None,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        b = MenuToolbarButton(self, id, title, bitmap, bitmap_disabled, kind)
        self.Append(b)
        return b


class MenuToolbarPage(ToolbarPage):
    def __init__(self, parent_bar, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        ToolbarPage.__init__(self, parent_bar, title, kind)

    def create_panel(self, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        m = MenuToolbarPanel(self, title, kind)
        return m


class MenuToolbarBar(ToolbarBar, wx.MenuBar):
    def __init__(self, parent, gui_style):
        wx.MenuBar.__init__(self)
        ToolbarBar.__init__(self, parent, gui_style)

    def create_page(self, title, kind):
        return MenuToolbarPage(self, title, kind)

    def update_bar(self, obj):
        obj.SetMenuBar(self.bar)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY  ; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

# Pytigon - wxpython and django application framework

# author: "Slawomir Cholaj (slawomir.cholaj@gmail.com)"
# license: "LGPL 3.0"

__version__ = "0.250214"



"""This module extends pytigon_lib.html module of additional class designed to implement widgets in the form."""

from urllib.parse import unquote
from base64 import b64decode

from pytigon_lib.schhtml.tags.table_tags import TableTag  # , TdTag
from pytigon_lib.schhtml.basehtmltags import (
    BaseHtmlElemParser,
    register_tag_map,
    register_tag_preprocess_map,
)
from pytigon_lib.schhtml.htmltools import superstrip
from pytigon_lib.schparser.html_parsers import Td
from pytigon_lib.schtools.tools import is_null
from pytigon_lib.schtools.schhtmlgen import make_start_tag
import wx

import pytigon_gui.guictrl.ctrl as schctrl


_ATTRIBUTES = (
    "name",
    "href",
    "label",
    "target",
    "valuetype",
    "defaultvalue",
    "length",
    "maxlength",
    "readonly",
    "src",
    "param",
    "fields",
    "hidden",
    "checked",
    "style",
    "id",
    "onload",
)


class TreeList(object):
    def __init__(self):
        self.children = []

    def append_to_tree(self, elem):
        self.children.append(elem)

    def get_list(self):
        return self.children


class TreeUl(BaseHtmlElemParser):
    def __init__(self, parent, parser, tag, attrs):
        BaseHtmlElemParser.__init__(self, parent, parser, tag, attrs)
        self.child_tags += ["li"]
        self.children = []

    def close(self):
        self.parent.append_to_tree(["", self.children, self.attrs])

    def append_to_tree(self, elem):
        self.children.append(elem)

    def handle_starttag(
        self,
        parser,
        tag,
        attrs,
    ):
        if tag == "li":
            return TreeLi(self, parser, tag, attrs)
        else:
            return None


class TreeLi(BaseHtmlElemParser):
    def __init__(
        self,
        parent,
        parser,
        tag,
        attrs,
    ):
        BaseHtmlElemParser.__init__(self, parent, parser, tag, attrs)
        self.child = None

    def close(self):
        if self.child:
            self.child[0] = superstrip("".join(self.data))
            self.child[2] = self.attrs
            self.parent.append_to_tree(self.child)
        else:
            self.parent.append_to_tree([superstrip("".join(self.data)), [], self.attrs])

    def append_to_tree(self, elem):
        self.child = elem

    def handle_starttag(self, parser, tag, attrs):
        if tag == "ul":
            return TreeUl(self, parser, tag, attrs)
        elif tag == "a" or tag == "ctrl-button":
            self.attrs["href"] = attrs["href"]
        else:
            return None


class Data(BaseHtmlElemParser):
    def __init__(self, parent, parser, tag, attrs):
        BaseHtmlElemParser.__init__(self, parent, parser, tag, attrs)
        self.data = None

    def close(self):
        if self.data:
            self.parent.data2 = b64decode(self.data.encode("utf-8")).decode("utf-8")
        else:
            self.parent.data2 = ""

    def handle_data(self, data):
        if self.data:
            self.data += data.rstrip(" \n")
        else:
            self.data = data.rstrip(" \n")


class OptionTag(BaseHtmlElemParser):
    def __init__(self, parent, parser, tag, attrs):
        BaseHtmlElemParser.__init__(self, parent, parser, tag, attrs)

    def close(self):
        if self.data and len(self.data) > 0:
            data = "".join(self.data).replace("\n", "").replace("\r", "").strip()
            # if 'selected' in self.attrs:
            #    data = '!!' + data
            # if 'value' in self.attrs:
            #    value = self.attrs['value']
            #    data = value + ':' + data
            # else:
            #    data = ':' + data
            self.parent.tdata.append([Td(data, self.attrs)])


class CompositeChildTag(BaseHtmlElemParser):
    """Class handle ctrlcomposite tag"""

    def __init__(self, parent, parser, tag, attrs):
        """Constructor

        Args:
            parent - parent tag
            parser - html parser object
            tag - tag
            attrs - attrs
        """
        BaseHtmlElemParser.__init__(self, parent, parser, tag, attrs)
        self.composite_data = {}
        self.composite_data["tag"] = tag
        self.composite_data["attrs"] = attrs
        self.composite_data["data"] = None
        self.composite_data["children"] = []

    def close(self):
        if self.data:
            self.composite_data["data"] = self.data
        if type(self.parent) == CompositeChildTag:
            self.parent.composite_data["children"].append(self.composite_data)
        elif type(self.parent).__name__ == "CtrlTag":
            self.parent.data2.append(self.composite_data)

    def handle_starttag(
        self,
        parser,
        tag,
        attrs,
    ):
        return CompositeChildTag(self, parser, tag, attrs)


class CtrlTag(TableTag):
    """Class which handle all tags startings with 'ctrl-'"""

    def __init__(self, parent, parser, tag, attrs):
        """Constructor

        Args:
            parent - parent tag
            parser - html parser object
            tag - tag
            attrs - attrs
        """

        TableTag.__init__(self, parent, parser, tag, attrs)

        self.list = None
        self.child_tags += ["ul", "li", "data", "option", "div"]
        self.subtab = True
        self.kwargs = {}
        self.classObj = None
        self.data2 = None
        self.obj = None
        self.width_best = -1
        self.width_min = -1
        self.width_max = -1
        self.tdata = []

        if self.parser.parse_only:
            return
        self._calc_relative_size()

        if self.tag == "ctrl-checkbox":
            pass

        if self.tag == "ctrl-composite":
            self.data2 = []

    def append_to_tree(self, elem):
        if not self.list:
            self.list = TreeList()
        self.list.append_to_tree(elem)

    def _calc_relative_size(self):
        if "width" in self.attrs:
            (width, height2) = self.get_parent_window().GetSize()
            if width == 1 and height2 == 1:
                (width, height2) = (
                    self.get_parent_window().GetParent().GetParent().GetSize()
                )
            if self.parser.dc.width > 0 and self.parser.dc.width < width:
                width = self.parser.dc.width
            m = self._get_parent_pseudo_margins()
            width = (width - m[0]) - m[1]
            self.width = self._norm_sizes([self.attrs["width"]], width)[0]

        if "height" in self.attrs:
            (width2, height) = self.get_parent_window().GetSize()
            if width2 == 1 and height == 1:
                (width2, height) = (
                    self.get_parent_window().GetParent().GetParent().GetSize()
                )

            m = self._get_parent_pseudo_margins()
            height = (height - m[2]) - m[3] - 3
            # if False:
            #    if '99.9%' in self.attrs['height']:
            #        height -= wx.Button.GetDefaultSize()[1]
            #    if '99.8%' in self.attrs['height']:
            #        height += m[2]+m[3]
            #    if '99.7%' in self.attrs['height']:
            #        height += m[2]+m[3]-wx.Button.GetDefaultSize()[1]
            # if '100%' in self.attrs['height']:
            #    if self.parent and hasattr(self.parent,'y'):
            #        height -= self.parent.y - +m[3]
            #    else:
            #        height += m[3]

            self.height = self._norm_sizes([self.attrs["height"]], height)[0]
            # if self.attrs["height"].strip() == "100%":
            #    self.height -= self.get_context()["top"]

    def get_context(self):
        if self.parent and hasattr(self.parent, "y"):
            _y = self.parent.y + 3
        else:
            _y = 1

        context = {
            "button_size_dx": wx.Button.GetDefaultSize()[0] + 14,
            "button_size_dy": wx.Button.GetDefaultSize()[1] + 14,
            "top": _y,
        }
        return context

    def handle_data(self, data):
        if self.tag == "ctrl-checkbox":
            pass
        return BaseHtmlElemParser.handle_data(self, data)

    def handle_starttag(self, parser, tag, attrs):
        if self.tag == "ctrl-checkbox":
            pass

        if self.tag == "ctrl-composite":
            return CompositeChildTag(self, parser, tag, attrs)

        if tag[:3] + "*" in self.child_tags:
            return self.class_from_tag_name(tag[:3] + "*")(self, parser, tag, attrs)
        elif tag == "ul":
            return TreeUl(self, parser, tag, attrs)
        elif tag == "li":
            return TreeLi(self, parser, tag, attrs)
        elif tag == "data":
            return Data(self, parser, tag, attrs)
        elif tag == "option":
            return OptionTag(self, parser, tag, attrs)
        elif tag in self.child_tags:
            o = self.class_from_tag_name(tag)
            if o:
                return o(self, parser, tag, attrs)
            else:
                return None
        else:
            return None

    def calc_col_sizes(self):
        sizes = self.calc_width()
        if sizes[0] >= 0:
            self.width = sizes[0] + self.extra_space[0] + self.extra_space[1]
        else:
            self.width = 400

    def get_height(self):
        height = self.calc_height()
        if height >= 0:
            self.height = height + self.extra_space[2] + self.extra_space[3]
        else:
            self.height = 200
        return self.height

    def calc_width(self):
        d = self.extra_space[0] + self.extra_space[1]
        if self.width < 0:
            if self.obj:
                (self.width_best, self.height_best) = self.obj.GetBestSize()
                (self.width_min, height_min) = self.obj.GetMinSize()
                (self.width_max, height_max) = self.obj.GetMaxSize()
                return (self.width_best + d, self.width_min + d, self.width_max + d)
            else:
                return (d, d, d)
        else:
            return (self.width, self.width, self.width)

    def calc_height(self):
        height = -1
        if self.height >= 0:
            return self.height
        if self.obj:
            (self.width_best, self.height_best) = self.obj.GetBestSize()
        else:
            self.height_best = 30
        self.height = self.height_best + self.extra_space[2] + self.extra_space[3]
        return self.height

    def _append_to_tdata(self):
        if len(self.tr_list) > 0:
            for tr in self.tr_list:
                row = []
                for td in tr:
                    td_obj = Td(td.to_txt(), td.attrs, td.to_obj_tab())
                    row.append(td_obj)
                self.tdata.append(row)

    def close(self):
        self._append_to_tdata()
        if self.parser.parse_only:
            if hasattr(self.parser, "register_tdata"):
                self.parser.register_tdata(self.tdata, self.tag, self.attrs)
        else:
            self.handle_ctrl(self.tag)
        return TableTag.close(self)

    def render(self, dc):
        if dc.rec:
            if self.obj:
                self.obj.SetSize(
                    int(dc.x + self.extra_space[0]),
                    int(dc.y + self.extra_space[2]),
                    int((self.width - self.extra_space[0]) - self.extra_space[1]),
                    int((self.height - self.extra_space[2]) - self.extra_space[3]),
                )
                x2 = self.obj.GetSize()
        return (self.height, False)

    def draw_atom(self, dc, style, x, y, dx, dy):
        dc2 = dc.subdc(x, y, self.width, self.height)
        cont = True
        while cont:
            (dy, cont) = self.render(dc2)
            dc2.y += dy

    def get_parent_window(self):
        return self.parser.get_parent_window()

    def handle_ctrl(self, tag):
        value = None
        width = -1
        height = -1
        if "name" in self.attrs:
            name_reg = is_null(self.attrs["name"], "")
            parent = self.parent
            while parent:
                if hasattr(parent, "reg_field"):
                    name_reg = parent.reg_field(self.attrs["name"])
                    break
                parent = parent.parent
            name = name_reg
        else:
            name = tag[4:]
        self.attrs["name"] = name
        parent = self.get_parent_window()
        if parent:
            gparent = parent.GetParent()
            tmp = name
            i = 1
            while gparent.test_ctrl(tmp):
                tmp = name + "__" + str(i)
                i += 1
            name = tmp
        for atrybut in _ATTRIBUTES:
            if atrybut in self.attrs:
                if self.attrs[atrybut] == None:
                    self.kwargs[atrybut] = ""
                else:
                    self.kwargs[atrybut] = self.attrs[atrybut]

        valuetype = "data"
        if "valuetype" in self.attrs:
            if self.attrs["valuetype"] == "ref":
                valuetype = "ref"
            if self.attrs["valuetype"] == "str":
                valuetype = "str"
        if "value" in self.attrs:
            if valuetype == "str":
                value = self.attrs["value"]
                value = unquote(value)
            else:
                try:
                    value = eval(self.attrs["value"])
                except:
                    print("Value error:", self.attrs["value"])
                    value = None
        else:
            value = None
        if "strvalue" in self.attrs:
            value = self.attrs["strvalue"]

        self.kwargs["size"] = wx.Size(
            int((self.width - self.extra_space[0]) - self.extra_space[1]),
            int((self.height - self.extra_space[2]) - self.extra_space[3]),
        )
        try:
            if tag.upper().startswith("CTRL-"):
                self.classObj = getattr(schctrl, tag[5:].upper())
            # elif tag.upper().startswith('CTRL'):
            #    self.classObj = getattr(schctrl, tag[4:].upper())
            elif tag.startswith("_"):
                self.classObj = getattr(schctrl, tag[1:].upper())
            else:
                self.classObj = getattr(schctrl, tag.upper())
        except:
            return True

        if len(self.tdata) > 0:
            self.kwargs["tdata"] = self.tdata
        self.kwargs["param"] = dict(self.attrs)
        self.kwargs["param"]["table_lp"] = str(self.parser.table_lp)
        self.kwargs["param"]["tag"] = tag.lower()
        if self.data2 != None:
            self.kwargs["param"]["data"] = is_null(self.data2, "")
        else:
            self.kwargs["param"]["data"] = "".join(self.data)
        if self.list:
            l = self.list.get_list()
            self.kwargs["ldata"] = l

        obj = None
        if parent:
            gparent = parent.GetParent()

            obj = gparent.pop_ctrl(name)
            if obj:
                if parent.update_controls:
                    if obj and hasattr(obj, "process_refr_data"):
                        obj.process_refr_data(**self.kwargs)
                    elif not (hasattr(obj, "is_ctrl_block") and obj.is_ctrl_block()):
                        obj.SetValue(value)
                parent.append_ctrl(obj)
            else:
                obj = self.classObj(parent, **self.kwargs)
                if not obj:
                    print("ERROR:", self.classObj)
                obj.set_unique_name(name)
                if value != None and (valuetype == "data" or valuetype == "str"):
                    obj.SetValue(value)
                obj.after_create()
                parent.append_ctrl(obj)
            self.obj = obj


register_tag_map("ctr*", CtrlTag)


class ComponentTag(CtrlTag):
    def __init__(self, parent, parser, tag, attrs):
        super().__init__(parent, parser, tag, attrs)
        self.data.append(make_start_tag(attrs["_tag"], attrs))
        self.count = 1
        self._close_tag = None

    def close(self):
        # self.count -= 1
        # if self.count < 0:

        if "window-width" in self.attrs:
            self.attrs["width"] = self.attrs["window-width"]

        if "window-height" in self.attrs:
            self.attrs["height"] = self.attrs["window-height"]

        return super().close()

    def handle_starttag(
        self,
        parser,
        tag,
        attrs,
    ):
        if "_tag" in attrs:
            tag2 = attrs["_tag"]
        else:
            tag2 = tag
        self.data.append(make_start_tag(tag2, attrs))
        if not self._close_tag:
            self._close_tag = self.close_tag
        self.count += 1
        return self

    def handle_endtag(self, tag):
        self.data.append("</" + tag + ">")
        self.count -= 1
        if self.count == 0:
            if self._close_tag:
                self.close_tag = self._close_tag
            return super().handle_endtag(tag)
        else:
            return self


register_tag_map("_component", ComponentTag)


def table_to_ctrltab(parent, attrs):
    if "class" in attrs:
        if "tabsort" in attrs["class"]:
            attrs["minheight"] = "60px"
            return ("ctrl-table", attrs)
        if "listctrl" in attrs["class"]:
            return ("ctrl-list", attrs)
    return ("table", attrs)


register_tag_preprocess_map("table", table_to_ctrltab)


SCHTYPE_MAP = {
    "FloatField": "float",
    "IntegerField": "num",
    "DecimalField": "amount",
    "TimeField": "time",
    "DateTimeField": "datetimepicker",
}


def input_to_ctrltab(parent, attrs):
    global SCHTYPE_MAP
    attrs_ret = {}
    ret = "text"
    upload = False
    if "type" in attrs:
        type = attrs["type"]
    else:
        type = "text"
    if "schtype" in attrs:
        schtype = attrs["schtype"]
    else:
        schtype = "text"
    if "class" in attrs:
        classname = attrs["class"]
    else:
        classname = "text"
    if schtype in SCHTYPE_MAP:
        ret = SCHTYPE_MAP[schtype]
    elif type == "text":
        if classname == "dateinput":
            ret = "datepicker"
        elif classname == "datetimeinput":
            ret = "datetimepicker"
        else:
            ret = "text"
    elif type == "button":
        ret = "button"
    elif type == "email":
        ret = "masktext"
        attrs_ret = {"src": "!EMAIL", "width": "250", "valuetype": "str"}
        if "value" in attrs:
            attrs_ret["value"] = attrs["value"]
    elif type == "hidden":
        ret = "text"
        attrs_ret = {"hidden": "1", "width": "0", "height": "0"}
    elif type == "password":
        ret = "password"
    elif type == "radio":
        ret = "radiobutton"
        if "checked" in attrs:
            attrs_ret["checked"] = "1"
        if "value" in attrs:
            attrs_ret["value"] = attrs["value"]
            attrs_ret["valuetype"] = "str"
    elif type == "checkbox":
        ret = "checkbox"
        if "checked" in attrs:
            attrs_ret["checked"] = "1"
        if "value" in attrs:
            attrs_ret["value"] = attrs["value"]
            attrs_ret["valuetype"] = "str"
    elif type == "file":
        ret = "filebrowsebutton"
        upload = True
    elif type == "submit":
        ret = "button"
        if "target" in attrs:
            target = attrs["target"]
        else:
            target = "_parent_refr"
        attrs_ret = {
            "defaultvalue": "1",
            "param": "post",
            "target": target,
            "id": "wx.ID_OK",
        }
        href = "."
        fields = None
        p = parent
        while p:
            if hasattr(p, "gethref"):
                href = p.gethref()
                fields = p.get_fields()
                if p.get_upload():
                    attrs_ret["valuetype"] = "upload"
                break
            p = p.parent
        if href == None:
            href = ""
        attrs_ret["href"] = href
        if fields:
            attrs_ret["fields"] = fields
    if "size" in attrs and type in ("text", "email"):
        width_text = 10 * int(is_null(attrs["size"], 20))
        if width_text > 400:
            width_text = 8 * int(is_null(attrs["size"], 20))
        if width_text > 480:
            width_text = 480
        attrs_ret["width"] = str(width_text)
    if "maxlength" in attrs and type in ("text",):
        attrs_ret["maxlength"] = str(is_null(attrs["maxlength"], "80"))
    elif "max_length" in attrs and type in ("text",):
        attrs_ret["maxlength"] = str(is_null(attrs["max_length"], "80"))
    else:
        if "size" in attrs:
            attrs_ret["maxlength"] = str(int(is_null(attrs["size"], "80")))
    if "value" in attrs and type in ("text", "hidden", "checkbox"):
        attrs_ret["value"] = is_null(attrs["value"], "")
        attrs_ret["valuetype"] = "str"
    if "value" in attrs and type in ("button", "submit"):
        attrs_ret["label"] = is_null(attrs["value"], "")
    if "title" in attrs and type in "checkbox":
        attrs_ret["label"] = is_null(attrs["title"], "")
    if "name" in attrs:
        p = parent
        while p:
            if p.form_obj:
                if upload:
                    p.form_obj.set_upload(upload)
                break
            p = p.parent
        attrs_ret["name"] = attrs["name"]
        if ret == "text":
            if "date_" in attrs["name"]:
                ret = "datepicker"
    for attr in attrs:
        if not attr in attrs_ret and attr != "value":
            attrs_ret[attr] = attrs[attr]
    return ("ctrl-" + ret, attrs_ret)


register_tag_preprocess_map("input", input_to_ctrltab)


def textarea_to_ctrltab(parent, attrs):
    attrs_ret = {}
    if "cols" in attrs:
        attrs_ret["cols"] = attrs["cols"]
    if "rows" in attrs:
        attrs_ret["rows"] = attrs["rows"]
    if "src" in attrs:
        attrs_ret["src"] = attrs["src"]
    if "name" in attrs:
        attrs_ret["name"] = attrs["name"]
    return ("ctrl-textarea", attrs_ret)


register_tag_preprocess_map("textarea", textarea_to_ctrltab)


def select_to_ctrltab(parent, attrs):
    return ("ctrl-select", attrs)


register_tag_preprocess_map("select", select_to_ctrltab)


def a_to_button(parent, attrs):
    if "class" in attrs:
        if (
            "popup" in attrs["class"]
            or "button" in attrs["class"]
            or "btn" in attrs["class"]
            or "schbtn" in attrs["class"]
        ):
            if "schbtn" in attrs["class"]:
                try:
                    if parent.parent.parent.tag == "ctrl-table":
                        return ("ctrl-button", attrs)
                except:
                    return ("a", attrs)
            else:
                try:
                    if parent.parent.parent.tag == "ctrl-table":
                        return ("a", attrs)
                except:
                    pass
                return ("ctrl-button", attrs)
    return ("a", attrs)


register_tag_preprocess_map("a", a_to_button)


HIDDEN_DIVS = [
    "ajax-region",
    "td_information",
    "td_action",
]


def div_convert(parent, attrs):
    if "class" in attrs:
        if "popup" in attrs["class"]:
            return ("div", attrs)
        if "form-group" in attrs["class"]:
            return ("tr", attrs)
        if "controls" in attrs["class"]:
            return ("td", attrs)
        if "tree" in attrs["class"]:
            return ("ctrl-tree", attrs)
        if "select2" in attrs["class"]:
            return ("ctrl-composite", attrs)
        if "checkbox" in attrs["class"]:
            obj = parent.parent.handle_starttag(parent.parser, "th", {})
            if obj:
                obj.handle_endtag("th")
            return ("th", attrs)
        for div in HIDDEN_DIVS:
            if div in attrs["class"]:
                return ("none", attrs)
    return ("div", attrs)


register_tag_preprocess_map("div", div_convert)


def label_to_th(parent, attrs):
    return ("th", attrs)


register_tag_preprocess_map("label", label_to_th)


def error_span_to_error(parent, attrs):
    if "class" in attrs and attrs["class"] == "help-block":
        return ("ctrl-errorlist", attrs)
    else:
        return ("span", attrs)


register_tag_preprocess_map("span", error_span_to_error)


def error_p_to_error(parent, attrs):
    if "class" in attrs and attrs["class"] == "help-block":
        return ("comment", attrs)
    else:
        return ("p", attrs)


register_tag_preprocess_map("p", error_p_to_error)


def ul_convert(parent, attrs):
    if "class" in attrs:
        if "root" in attrs["class"]:
            return ("ldata", attrs)
        if attrs["class"] == "errorlist":
            return ("ctrl-errorlist", attrs)
    return ("ul", attrs)


register_tag_preprocess_map("ul", ul_convert)


def component_convert(parent, attrs):
    return ("_component", attrs)


register_tag_preprocess_map("*-*", component_convert, "ctr*")



import collections
import wx

from pytigon_gui.guilib.image import bitmap_from_href

_ = wx.GetTranslation


class SchGridPanel(wx.Panel):
    """Panel for grid widget with toolbar"""

    def __init__(self, *args, **argv):
        self.grid = None
        self.icon_size = 2
        wx.Panel.__init__(self, *args, **argv)
        self.vertical = False
        self._bitmaps = {
            "edit": "wx.ART_FILE_OPEN",
            "edit_inline": "wx.ART_FILE_OPEN",
            "delete": "wx.ART_DELETE",
            "view_row": "wx.ART_INFORMATION",
        }
        self._menu_buttons = collections.OrderedDict()

    def set_bitmap(self, action, path):
        self._bitmaps[action] = path

    def set_bitmaps(self, action_dict):
        for key in action_dict:
            self.set_bitmap(key, action_dict[key])

    def set_vertical(self, enable=True):
        self.vertical = enable

    def _get_bmp(self, id_str):
        if id_str in self._bitmaps:
            value = self._bitmaps[id_str]
            return bitmap_from_href(value, self.icon_size)
        return wx.ArtProvider.GetBitmap(wx.ART_MISSING_IMAGE, wx.ART_TOOLBAR, (32, 32))

    def _add_action(self, action):
        test = False
        if "name" in action:
            name = action["name"]
            label = action["data"]
            if "title" in action:
                title = action["title"]
                if not label:
                    label = title
            else:
                title = ""
            if (
                not name in ("insert", "edit", "delete", "new", "get_row", "view_row")
                and not name in self.commands
            ):
                if name in self._bitmaps:
                    b = self._get_bmp(name)
                else:
                    if "src" in action:
                        b = bitmap_from_href(action["src"])
                    else:
                        b = bitmap_from_href("fa://fa-chevron-right?size=1")
                if "(" in label and ")" in label:
                    x = label.split("(")
                    label1 = x[0]
                    label2 = x[1].split(")")[0]
                    if not label2 in self._menu_buttons:
                        self.toolbar.AddTool(
                            self.lp2,
                            label2,
                            b,
                            wx.NullBitmap,
                            wx.ITEM_NORMAL,
                            label2,
                            title,
                        )
                        self._menu_buttons[label2] = []
                        self.lp2 += 1
                    tab_menu = self._menu_buttons[label2]
                    tab_menu.append((self.lp, label1))
                else:
                    self.toolbar.AddTool(
                        self.lp, label, b, wx.NullBitmap, wx.ITEM_NORMAL, label, title
                    )
                self.commands.append(name)
                test = True
                self.lp += 1
        return test

    def create_toolbar(self, grid):
        self.GetParent().signal_from_child(self, "set_bitmap_list")
        self.grid = grid
        self.spanel = wx.ScrolledWindow(
            self, style=wx.VSCROLL if self.vertical else wx.HSCROLL
        )
        self.commands = []
        self.lp = 101
        self.lp2 = 201
        (standard, akcje) = grid.get_action_list()
        if standard or akcje and len(akcje) > 0:
            if self.vertical:
                self.toolbar = wx.ToolBar(
                    self.spanel, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_VERTICAL
                )
            else:
                self.toolbar = wx.ToolBar(
                    self.spanel, -1, wx.DefaultPosition, wx.DefaultSize, 0
                )
            tsize = (32, 32)
            self.toolbar.SetToolBitmapSize(tsize)
            if standard:
                self.toolbar.AddTool(
                    self.lp,
                    _("Edit"),
                    self._get_bmp("edit"),
                    wx.NullBitmap,
                    wx.ITEM_NORMAL,
                    _("Edit"),
                    _("Edit"),
                )
                self.toolbar.AddTool(
                    self.lp + 1,
                    _("Delete"),
                    self._get_bmp("delete"),
                    wx.NullBitmap,
                    wx.ITEM_NORMAL,
                    _("Delete"),
                    _("Delete"),
                )
                self.toolbar.AddTool(
                    self.lp + 2,
                    _("view_row"),
                    self._get_bmp("view_row"),
                    wx.NullBitmap,
                    wx.ITEM_NORMAL,
                    _("View row"),
                    _("View row"),
                )
                self.toolbar.AddSeparator()
                self.commands.append("edit")
                self.commands.append("delete")
                self.commands.append("view_row")
                self.lp += 3
            if akcje:
                for action in akcje:
                    self._add_action(action)
            self.toolbar.Realize()
            self.toolbar.Bind(wx.EVT_TOOL, self.on_tool_click)
            self.Bind(wx.EVT_MENU, self.on_tool_click)

            self.toolbar.SetSize(self.toolbar.GetBestSize())
            if self.vertical:
                self.spanel.SetScrollRate(0, 20)
            else:
                self.spanel.SetScrollRate(20, 0)
            self.spanel.SetVirtualSize(self.toolbar.GetSize())
        grid.set_panel(self)
        self.Bind(wx.EVT_SIZE, self.on_size)
        self.on_size()

    def on_size(self, event=None):
        if event:
            panel_size = event.GetSize()
        else:
            panel_size = self.GetSize()
        toolbar_size = self.toolbar.GetSize()
        if self.vertical:
            if toolbar_size[1] >= panel_size[1]:
                dx = toolbar_size[0] + wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X)
            else:
                dx = toolbar_size[0]
            self.spanel.SetRect(wx.Rect(0, 0, dx, panel_size[1]))
            self.grid.SetRect(
                wx.Rect(dx + 2, 0, (panel_size[1] - dx) - 2, panel_size[1])
            )
        else:
            if toolbar_size[0] >= panel_size[0]:
                dy = toolbar_size[1] + wx.SystemSettings.GetMetric(wx.SYS_HSCROLL_Y)
            else:
                dy = toolbar_size[1]
            self.spanel.SetRect(wx.Rect(0, 0, panel_size[0], dy))
            self.grid.SetRect(
                wx.Rect(0, dy + 2, panel_size[0], (panel_size[1] - dy) - 2)
            )
        if event:
            event.Skip()

    def on_tool_click(self, event):
        id = event.GetId()
        if id > 200:
            id2 = id - 201
            tab_menu = list(self._menu_buttons.values())[id2]
            menu = wx.Menu()
            for pos in tab_menu:
                menu.Append(pos[0], pos[1])
            self.PopupMenu(menu)
            menu.Destroy()
        else:
            self.grid.action(self.commands[id - 101])

    def refresh(self, row):
        if self.grid.GetTable().GetNumberRows() > 0:
            akcje = self.grid.get_action_list(row)[1]
            akcje_dict = {}
            test = False
            if akcje:
                for action in akcje:
                    if not "name" in action:
                        continue
                    akcje_dict[action["name"]] = akcje
                    if self._add_action(action):
                        test = True
            if test:
                self.toolbar.Realize()
                self.toolbar.SetSize(self.toolbar.GetBestSize())
            i = 101
            for command in self.commands:
                if command in akcje_dict:
                    self.toolbar.EnableTool(i, True)
                else:
                    self.toolbar.EnableTool(i, False)
                i = i + 1



import wx

import wx.lib.agw
import wx.lib.agw.aui as aui
import wx.lib.agw.aui.aui_utilities

from pytigon_gui.guilib.events import *
from pytigon_gui.toolbar.basetoolbar import (
    ToolbarBar,
    ToolbarPage,
    ToolbarPanel,
    ToolbarButton,
)


_ = wx.GetTranslation


class SchAuiToolBarArt(aui.AuiDefaultToolBarArt):
    def __init__(self):
        aui.AuiDefaultToolBarArt.__init__(self)
        self._base_colour = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND)

    def draw_background(self, dc, wnd, _rect, horizontal=True):
        rect = wx.Rect(*_rect)
        start_colour = self._base_colour
        end_colour = self._base_colour
        reflex_colour = aui.StepColour(self._base_colour, 95)
        dc.GradientFillLinear(
            rect, start_colour, end_colour, (horizontal and [wx.SOUTH] or [wx.EAST])[0]
        )
        left = rect.GetLeft()
        right = rect.GetRight()
        top = rect.GetTop()
        bottom = rect.GetBottom()
        dc.SetPen(wx.Pen(reflex_colour))
        if horizontal:
            dc.DrawLine(left, bottom, right + 1, bottom)
        else:
            dc.DrawLine(right, top, right, bottom + 1)


class GenericToolbarButton(ToolbarButton):
    def __init__(
        self,
        parent_panel,
        id,
        title,
        bitmap,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        ToolbarButton.__init__(
            self, parent_panel, id, title, bitmap, bitmap_disabled, kind
        )


class GenericToolbarPanel(ToolbarPanel):
    def __init__(self, parent_page, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        ToolbarPanel.__init__(self, parent_page, title, kind)

        if len(parent_page.panels) > 0:
            parent_page.AddStretchSpacer()

    def _append(self, b):
        item = None
        if self.kind in (
            ToolbarPanel.TYPE_PANEL_TOOLBAR,
            ToolbarPanel.TYPE_PANEL_BUTTONBAR,
        ):
            if b.kind == ToolbarButton.TYPE_SIMPLE:
                item = self.parent_page.AddSimpleTool(b.id, b.title, b.bitmap, b.title)
            elif b.kind == ToolbarButton.TYPE_DROPDOWN:
                item = self.parent_page.AddSimpleTool(b.id, b.title, b.bitmap, b.title)
            elif b.kind == ToolbarButton.TYPE_HYBRID:
                item = self.parent_page.AddSimpleTool(b.id, b.title, b.bitmap, b.title)
            elif b.kind == ToolbarButton.TYPE_TOOGLE:
                item = self.parent_page.AddToggleTool(b.id, b.title, b.bitmap)
            elif b.kind == ToolbarButton.TYPE_PANEL:
                pass
            elif b.kind == ToolbarButton.TYPE_SEPARATOR:
                self.parent_page.AddSeparator()
        else:
            pass

    def create_button(
        self,
        id,
        title,
        bitmap=None,
        bitmap_disabled=None,
        kind=ToolbarButton.TYPE_SIMPLE,
    ):
        b = GenericToolbarButton(self, id, title, bitmap, bitmap_disabled, kind)
        self._append(b)
        return b

    def add_separator(self):
        b = GenericToolbarButton(
            self, 0, "", None, None, kind=ToolbarButton.TYPE_SEPARATOR
        )
        self._append(b)
        return b


class GenericToolbarPage(ToolbarPage, aui.AuiToolBar):
    def __init__(self, parent_bar, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        ToolbarPage.__init__(self, parent_bar, title, kind)
        agwStyle = aui.AUI_TB_DEFAULT_STYLE | aui.AUI_TB_HORZ_LAYOUT
        aui.AuiToolBar.__init__(
            self,
            parent_bar.parent,
            -1,
            wx.DefaultPosition,
            wx.DefaultSize,
            agwStyle=agwStyle,
        )

        self.SetToolBitmapSize(wx.Size(24, 24))
        if wx.Platform != "__WXMSW__":
            self.SetArtProvider(SchAuiToolBarArt())
        nr = len(parent_bar.pages) + 1

        pinfo = (
            parent_bar.parent._create_pane_info("tb" + str(nr), "Toolbar " + str(nr))
            .ToolbarPane()
            .Top()
            .LeftDockable(False)
            .RightDockable(False)
            .Row(1)
        )
        parent_bar.parent._mgr.AddPane(self, pinfo)

    def create_panel(self, title, kind=ToolbarPanel.TYPE_PANEL_TOOLBAR):
        return GenericToolbarPanel(self, title, kind)


class GenericToolbarBar(ToolbarBar):
    def __init__(self, parent, gui_style):
        ToolbarBar.__init__(self, parent, gui_style)

    def create_page(self, title, kind=ToolbarPage.TYPE_PAGE_NORMAL):
        return GenericToolbarPage(self, title, kind)

    def create(self):
        for bar in self.pages:
            bar.Realize()
            bar.SetSize(bar.GetBestSize())

    def update(self):
        pass

    def bind_ui(self, fun, id=wx.ID_ANY):
        self.parent.Bind(wx.EVT_UPDATE_UI, fun, id=id)

    def bind(self, fun, id=wx.ID_ANY, e=None):
        if e:
            self.parent.Bind(e, fun, id=id)
        else:
            self.parent.Bind(wx.EVT_MENU, fun, id=id)

    def bind_dropdown(self, fun, id):
        self.parent.Bind(wx.EVT_MENU, fun, id=id)

    def un_bind(self, id, e=None):
        if e:
            self.Unbind(e, id=id)
        else:
            self.Unbind(wx.EVT_MENU, id=id)



"""Module contains helper classes and functions for operate with images"""

import io
from io import BytesIO
from PIL import Image
from wx.svg import SVGimage

import wx

_ = wx.GetTranslation


SIZE_DEFAULT = -1
SIZE_SMALL = 0
SIZE_MEDIUM = 1
SIZE_LARGE = 2


# def get_app_icon():
#    icon = wx.EmptyIcon()
#    b = wx.ArtProvider.GetBitmap(wx.ART_FRAME_ICON, wx.ART_TOOLBAR, (32, 32))
#    icon.CopyFromBitmap(b)
#    return icon


def bitmap_to_pil(bitmap):
    """Convert wx.Bitmap object to PIL bitmap"""
    return image_to_pil(bitmap_to_image(bitmap))


def bitmap_to_image(bitmap):
    """Convert wx.Bitmap object to wx.Image object"""
    return wx.ImageFromBitmap(bitmap)


def pil_to_bitmap(pil):
    """Convert PIL bitmap to wx.Bitmap"""
    return image_to_bitmap(pil_to_image(pil))


def pil_to_image(pil, alpha=True):
    """Convert PIL bitmap to wx.Image"""
    if alpha:
        image = wx.Image(*pil.size)
        image.SetData(pil.convert("RGB").tobytes())
        image.SetAlphaBuffer(pil.convert("RGBA").tobytes()[3::4])
    else:
        image = wx.Image(pil.size[0], pil.size[1])
        new_image = pil.convert("RGB")
        data = new_image.tobytes()
        image.SetData(data)
    return image


def image_to_pil(image):
    """Convert wx.Image to PIL image"""
    pil = Image.new("RGB", (image.GetWidth(), image.GetHeight()))
    pil.fromstring(image.GetData())
    return pil


def image_to_bitmap(image):
    """Convert wx.Image to wx.Bitmap"""
    return image.ConvertToBitmap()


def bitmap_from_art_id(art_id, size):
    """Get bitmap from ArtProvider

    Args:
        art_id - Bitmap id in ArtProvider
        size
    """
    return wx.ArtProvider.GetBitmap(art_id, wx.ART_OTHER, size)


def bitmaps_from_art_id(art_id, size):
    """Get tuple with 2 bitmaps from ArtProvider, first is standard bitmap, second is first bitmap convertet
    to grayscale to emulate disabled elements

    Args:
        art_id - Bitmap id in ArtProvider
        size
    """
    bitmap = wx.ArtProvider.GetBitmap(art_id, wx.ART_OTHER, size)
    img = bitmap.ConvertToImage()
    bitmap_disabled = wx.Bitmap(img.ConvertToGreyscale().AdjustChannels(1, 1, 1, 0.3))
    return (bitmap, bitmap_disabled)


def bitmap_from_href(href, size_type=SIZE_DEFAULT):
    """Load resource and convert to bitmap

    Args:
        href - can be:
            - ArtProvider id - for example "wx.ART_COPY"
            - Font Awesome id, format: "fa://icon_name", for example "fa://bus"
            - path to image in local library, format: "png://image_name",
              for example: "png://mimetypes/text-x-script.png".
              Local library is located in path: "/appdata/media/size/"
            - path to image in web, for example: "http://www.pytigon.cloud/test.png"

        size
    """
    size_type2 = size_type
    if "?" in href:
        x = href.split("?")
        href2 = x[0]
        if "=" in x[1]:
            try:
                size_type2 = int(x[1].split("=")[1])
            except:
                size_type2 = 1
    else:
        href2 = href
    if size_type2 == SIZE_SMALL:
        icon_size = 16
    elif size_type2 == SIZE_MEDIUM:
        icon_size = 22
    else:
        icon_size = 32
    if href2[:3] == "wx.":
        bmp = wx.ArtProvider.GetBitmap(
            eval(href2), wx.ART_TOOLBAR, (icon_size, icon_size)
        )
    elif href.startswith("client://"):
        image = wx.Image(
            wx.GetApp().src_path
            + "/static/icons/%dx%d/" % (icon_size, icon_size)
            + href2[9:]
        )
        bmp = wx.Bitmap(image)
    elif href.startswith("png://"):
        image = wx.Image(
            wx.GetApp().src_path
            + "/static/icons/%dx%d/" % (icon_size, icon_size)
            + href2[6:]
        )
        bmp = wx.Bitmap(image)
    elif href.startswith("fa://"):
        if ".png" in href.lower():
            suffix = ""
        else:
            suffix = ".png"
        try:
            image = wx.Image(
                wx.GetApp().src_path
                + "/static/fonts/fork-awesome/fonts/%dx%d/" % (icon_size, icon_size)
                + href2[5:].replace("fa-", "")
                + suffix
            )
            image = image.AdjustChannels(1, 1, 1, 0.55)
            bmp = wx.Bitmap(image)
        except:
            print("Error, can't load image: ", href2)
            bmp = wx.Bitmap()
    elif href.startswith("data:image/svg+xml"):
        x = href.split(",", 1)
        svg_code = x[1]
        svg = SVGimage.CreateFromBytes(svg_code.encode("utf-8"))
        # _bmp = svg.ConvertToBitmap(scale=10, width=icon_size*10, height=icon_size*10)
        # image = wx.ImageFromBitmap(_bmp)
        # image = image.Scale(width=icon_size, height=icon_size, quality=wx.IMAGE_QUALITY_HIGH)
        # bmp = wx.Bitmap(image)
        bmp = svg.ConvertToBitmap(width=icon_size, height=icon_size)
    else:
        if href.lower().startswith("http"):
            http = wx.GetApp().get_http(None)
            response = http.get(None, str(href))
            if response.ret_code != 404:
                s = response.ptr()
                stream = BytesIO(s)
                bmp = wx.Bitmap(wx.Image(stream))
            else:
                bmp = wx.ArtProvider.GetBitmap(
                    wx.ART_MISSING_IMAGE, wx.ART_TOOLBAR, (32, 32)
                )
        else:
            bmp = wx.ArtProvider.GetBitmap(
                wx.ART_MISSING_IMAGE, wx.ART_TOOLBAR, (32, 32)
            )
    return bmp


class ArtProviderFromIcon(wx.ArtProvider):
    """Extended version of wx.ArtProvider - class loads bitmaps from local library"""

    def __init__(self):
        wx.ArtProvider.__init__(self)
        self.tab_16 = {}
        self.tab_22 = {}
        self.tab_32 = {}

        ids = open(wx.GetApp().src_path + "/static/icons/ids.txt")
        for line in ids:
            l = line.replace("\n", "").split(",")
            if len(l) > 1:
                image_path = l[1]
                i = "wx.ART_" + l[0]
                try:
                    wxid = eval(i)
                    self.tab_16[wxid] = [image_path, None]
                    self.tab_22[wxid] = [image_path, None]
                    self.tab_32[wxid] = [image_path, None]
                except:
                    pass
        ids.close()

    def CreateBitmap(self, artid2, client, size):
        artid = artid2.encode("utf-8")
        if size != 16 and size != 22 and size != 32:
            size = 32
        path = None
        if size == 16:
            if artid in self.tab_16:
                if self.tab_16[artid][1]:
                    return self.tab_16[artid][1]
                path = "png://" + self.tab_16[artid][0] + "?size=0"
        elif size == 22:
            if artid in self.tab_22:
                if self.tab_22[artid][1]:
                    return self.tab_22[artid][1]
                path = "png://" + self.tab_22[artid][0] + "?size=1"
        else:
            if artid in self.tab_32:
                if self.tab_32[artid][1]:
                    return self.tab_32[artid][1]
                path = "png://" + self.tab_32[artid][0] + "?size=2"
        try:
            bitmap = bitmap_from_href(path)
        except:
            bitmap = None
        if bitmap:
            if size == 16:
                self.tab_16[artid][1] = bitmap
            elif size == 22:
                self.tab_22[artid][1] = bitmap
            else:
                self.tab_32[artid][1] = bitmap
            return bitmap
        else:
            return wx.NullBitmap


class SchImage:
    """Class which extract bitmaps from image with multiple images aranged one behind other.
    All partial bitmaps should be the same size. Between bitmaps should be a gap with a 1 pixel width.

    images = SchImage("http://wwww.page.com", (32,32))
    image1 = images[0]
    image2 = images[2]
    """

    def __init__(self, address, one_image_size=(20, 20)):
        self.one_image_size = one_image_size
        try:
            http = wx.GetApp().http
            response = http.get(self, address)
            if response.ret_code != 200:
                print("I can't load image from:", address)
                self.bmp = wx.ArtProvider.GetBitmap(
                    wx.ART_MISSING_IMAGE, wx.ART_TOOLBAR, (32, 32)
                )
            else:
                str = response.ptr()
                stream = io.BytesIO(str)
                self.bmp = wx.Bitmap(wx.Image(stream))
        except:
            print(_("The exception while loading image from:"), address)
            self.bmp = wx.ArtProvider.GetBitmap(
                wx.ART_MISSING_IMAGE, wx.ART_TOOLBAR, (32, 32)
            )

    def __getitem__(self, key):
        rect = wx.Rect(
            (self.one_image_size[0] + 1) * int(key),
            0,
            self.one_image_size[0],
            self.one_image_size[1],
        )
        try:
            ret = self.bmp.GetSubBitmap(rect)
        except:
            ret = wx.ArtProvider.GetBitmap(
                wx.ART_MISSING_IMAGE, wx.ART_TOOLBAR, (32, 32)
            )
        return ret



import os
import sys
import wx

from pytigon_lib.schhttptools.httpclient import COOKIES

import pytigon_gui.guictrl.ctrl

from pytigon_gui.guiframe.baseframe import SchBaseFrame

from django.conf import settings

from pytigon.pytigon_request import init, request

_ = wx.GetTranslation


class SchBrowserFrame(SchBaseFrame):
    """
    This is main window of pytigon application
    """

    def __init__(
        self,
        parent,
        gui_style="tree(toolbar,statusbar)",
        id=-1,
        title="",
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.DEFAULT_FRAME_STYLE | wx.CLIP_CHILDREN | wx.WANTS_CHARS,
        name="MainWindow",
    ):
        self.gui_style = gui_style
        self.destroy_fun_tab = []
        self.idle_objects = []
        self.after_init = False
        self.desktop = None
        self._mgr = None
        self.toolbar_interface = None
        self.aTable = None
        self.ctrl = None

        SchBaseFrame.__init__(
            self, parent, id, gui_style, title, pos, size, style | wx.WANTS_CHARS, name
        )
        wx.GetApp().SetTopWindow(self)

        self.init_plugins()
        self.Bind(wx.EVT_IDLE, self.on_idle)
        self.Bind(wx.EVT_SIZE, self.on_size)

        init(os.environ["PRJ_NAME"], "auto", "anawa", user_agent="webviewembeded")
        start_request = request("/", None, user_agent="webviewembeded")

        self.ctrl = pytigon_gui.guictrl.ctrl.HTML2(
            self, name="schbrowser", size=self.GetClientSize()
        )
        self.ctrl.load_str(start_request.str(), "http://127.0.0.5/")

        if sys.platform != "win32":
            while not self.ctrl.page_loaded:
                wx.Yield()

        size = wx.GetApp().app_size
        wx.CallAfter(self.SetSize, (size[0], size[1]))
        wx.CallAfter(self.Show)

    def on_size(self, event):
        if event:
            if self.ctrl:
                self.ctrl.SetSize(event.GetSize())
            event.Skip()
        else:
            if self.ctrl:
                self.ctrl.SetSize(self.GetSize())

    def get_menu_bar(self):
        return None

    def on_idle(self, event):
        for obj in self.idle_objects:
            obj.on_idle()

        if not self.after_init:
            self.after_init = True
            app = wx.GetApp()
            if len(app.start_pages) > 0:

                def start_pages():
                    for page in app.start_pages:
                        url_page = page.split(";")
                        if len(url_page) == 2:
                            self._on_html(
                                _(url_page[0]) + "," + app.base_address + url_page[1]
                            )

                wx.CallAfter(start_pages)

        event.Skip()

    def set_acc_key_tab(self, win, tab):
        pass



import copy

import wx
import pytigon_lib.schtools.createparm as createparm

from .gridtable_base import SchGridTableBase


class DataSource(SchGridTableBase):
    def __init__(self, proxy):
        SchGridTableBase.__init__(self)
        self.proxy = proxy
        self.proxy.set_parent(self)
        self.last_page_id = -1
        self.pages_nr = [-1, -1, -1, -1, -1]
        self.pages = [None, None, None, None, None]
        self.last_page = None
        self.max_count = self.proxy.get_max_count()
        self.rec_count = self.max_count

        self.default_rec = self.proxy.get_default_rec()

        self.attrs = {}
        self.setparm = False

        self.can_append = 1
        self.read_only = False
        self.is_valid = True

        self.last_row_count = self.max_count + self.can_append

    def invalidate(self):
        self.is_valid = False

    def _scroll(self, end):
        for i in range(end, 4):
            (self.pages)[4 - i] = (self.pages)[3 - i]
            (self.pages_nr)[4 - i] = (self.pages_nr)[3 - i]

    def refr_count(self, count, storePos=0):
        self.rec_count = count - self.can_append - self.append_count
        return SchGridTableBase.refr_count(self, count, storePos)

    def refresh(self, storePos):
        # store pos
        # 0 - no store, cursor - pos 1
        # 1 - store, if out of range - pos 1
        # 2 - store, if out of range - pos = max
        self.commit()
        self.append_count = 0
        self.last_page_id = -1
        self.pages_nr = [-1, -1, -1, -1, -1]
        self.pages = [None, None, None, None, None]
        self.last_page = None
        self.max_count = self.proxy.get_max_count()
        self.rec_to_update = dict()
        self.rec_to_instert = dict()
        self.rec_to_delete = []
        self.refr_count(
            self.proxy.get_count() + self.can_append + self.append_count, storePos
        )
        self.GetView().ForceRefresh()

    def get_rec(self, nr_rec):
        if not self.is_valid:
            return None
        if nr_rec < self.rec_count or self.rec_count == self.max_count:
            strona = nr_rec // 256
            if strona == self.last_page_id:
                if self.last_page != None and nr_rec % 256 < len(self.last_page):
                    return (self.last_page)[nr_rec % 256]
                else:
                    return None
            else:
                if strona in self.pages_nr:
                    id = self.pages_nr.index(strona)
                    self.last_page = (self.pages)[id]
                    self.last_page_id = strona
                    self._scroll(4 - id)
                    (self.pages)[0] = self.last_page
                    (self.pages_nr)[0] = strona
                    return self.get_rec(nr_rec)
                else:
                    if self.setparm == False:
                        self.setparm = True
                        parm = createparm.create_parm(
                            self.GetView().address,
                            self.GetView().GetParent().get_parm_obj(),
                        )
                        if parm:
                            self.proxy.set_address_parm(parm[2])

                    if self.proxy.is_valid:
                        nowaStrona = self.proxy.get_page(strona)
                    else:
                        self.is_valid = False
                        nowaStrona = None

                    if nowaStrona != None:
                        self._scroll(0)
                        (self.pages_nr)[0] = strona
                        (self.pages)[0] = nowaStrona
                        self.last_page = (self.pages)[0]
                        self.last_page_id = strona

                        if len(
                            nowaStrona
                        ) < 256 and self.rec_count != strona * 256 + len(nowaStrona):
                            self.refr_count(
                                self.proxy.get_count()
                                + self.can_append
                                + self.append_count,
                                2,
                            )
                        return self.get_rec(nr_rec)
                    else:
                        self.refr_count(self.proxy.get_count(), 2)
                        return None
        else:
            return None

    def commit(self):
        if self.read_only:
            return

        test = False
        update = []

        if len(self.rec_to_instert) > 0 or len(self.rec_to_delete) > 0:
            test = True

        if test or len(self.rec_to_update) > 0:
            self.rec_count = self.rec_count + len(self.rec_to_instert)
            for rec in list(self.rec_to_update.keys()):
                update.append((self.rec_to_update)[rec])
            insert = []
            for rec in list(self.rec_to_instert.keys()):
                insert.append((self.rec_to_instert)[rec])
            delete = []
            for rec in self.rec_to_delete:
                delete.append(self.get_rec(rec)[0])
            self.proxy.sync_data(update, insert, delete)

            self.rec_to_update = dict()
            self.rec_to_instert = dict()
            self.rec_to_delete = []

    def data_changed(self):
        if self.read_only:
            return False
        if (
            len(list(self.rec_to_update.keys())) > 0
            or len(list(self.rec_to_instert.keys())) > 0
            or len(self.rec_to_delete) > 0
        ):
            return True
        else:
            return False

    def get_rec_count(self):
        return self.rec_count

    def _get_number_rows(self):
        return self.rec_count + self.can_append + self.append_count

    def get_action_list(self, row, col=None):
        return self.rec[row]

    def get_actions(self, row, col=None):
        rowlen = len(self.GetColNames())
        rec = self.get_rec(row)
        if rec and len(rec) > rowlen + 1:
            return rec[rowlen + 1]
        else:
            return dict()

    def sort(self, column, append):
        SchGridTableBase.sort(self, column, append)
        sort = ""
        test = 0
        if len(self.tabsort) > 0:
            for i in self.tabsort:
                if test == 0:
                    test = 1
                else:
                    sort = sort + ","

                if i < 0:
                    sort = sort + "-" + self.GetColNames()[-1 * i]
                else:
                    sort = sort + self.GetColNames()[i]

            self.proxy.set_parm("sort", sort)

        self.refresh(0)

    def auto_update(self, col_name, col_names, rec):
        """Return transformed row after current row is changed"""
        return self.proxy.auto_update(col_name, col_names, rec)

    def filter_cmp(self, pos, key):
        if self.filter_id >= 0:
            elem = pos[self.filter_id]
            if elem.__class__ == tuple:
                elem = elem[0]
            if str(elem).upper().startswith(key.upper()):
                return True
            else:
                return False
        else:
            return False

    def filter(self, key):
        self.key = key
        self.proxy.set_parm("value", key)
        self.refresh(0)

    def get_table_and_state(self):
        tableandstate = SchGridTableBase.get_table_and_state(self)
        tableandstate.append(self.proxy)
        tableandstate.append(
            (
                self.last_page_id,
                self.pages_nr,
                self.pages,
                self.last_page,
                self.max_count,
                self.rec_count,
            )
        )
        return tableandstate

    def duplicate_table_and_state(self):
        tableandstate = SchGridTableBase.duplicate_table_and_state(self)
        self.proxy = self.proxy.clone()
        self.last_page_id = -1
        self.pages_nr = [-1, -1, -1, -1, -1]
        self.pages = [None, None, None, None, None]
        self.last_page = None
        self.max_count = self.proxy.get_max_count()
        self.rec_count = self.max_count
        return tableandstate

    def set_table_and_state(self, tableandstate):
        oldtableandstate = self.get_table_and_state()
        tbl = tableandstate[1:]
        self.proxy = tbl[0]
        self.last_page_id = tbl[1][0]
        self.pages_nr = tbl[1][1]
        self.pages = tbl[1][2]
        self.last_page = tbl[1][3]
        self.max_count = tbl[1][4]
        self.rec_count = tbl[1][5]

        SchGridTableBase.set_table_and_state(self, tableandstate)

        return oldtableandstate

    def run_cmd(self, cmd, parm=None, option=0):
        x = self.make_exec_cmd(cmd, option)
        if parm:
            x.append(parm)
        ret = self.proxy.exec(x)
        return ret

    def GetRecAsStr(self, nrRec):
        return self.proxy.GetRecAsStr(nrRec)

    def GetColNames(self):
        return self.proxy.GetColNames()

    def GetAutoCols(self):
        return self.proxy.GetAutoCols()

    def GetColTypes(self):
        return self.proxy.GetColTypes()

    def GetColIcons(self):
        return self.proxy.GetColIcons()

    def GetAttr(self, row, col, kind):
        if not self.is_valid:
            return SchGridTableBase.GetAttr(self, row, col, kind)
        if (
            row < self.rec_count
            and (row not in self.rec_to_update)
            and (row not in self.rec_to_delete)
            and (not self._is_sel(row))
        ):
            rec = self.get_rec(row)
            if rec:
                value = rec[col + 1]
                if value.__class__ in (tuple, list):
                    if value[1] in self.attrs:
                        attr = self.attrs[value[1]]
                    else:
                        attr = wx.grid.GridCellAttr()
                        atrybuty = value[1].split(",")
                        style = atrybuty[0]
                        bgcolor = None
                        color = None
                        if len(atrybuty) > 1:
                            bgcolor = atrybuty[1]
                        if len(atrybuty) > 2:
                            color = atrybuty[2]

                        if bgcolor and len(bgcolor) > 1:
                            attr.SetBackgroundColour(bgcolor)
                        if color and len(color) > 1:
                            attr.SetTextColour(color)
                        if style and len(style) > 0:
                            if "s" in style:
                                font = self.GetView().GetDefaultCellFont()
                                font.SetWeight(wx.FONTWEIGHT_BOLD)
                                attr.SetFont(font)
                            if ">" in style:
                                attr.SetAlignment(wx.ALIGN_RIGHT, -1)
                            if "-" in style:
                                attr.SetAlignment(wx.ALIGN_CENTER, -1)
                        (self.attrs)[value[1]] = attr
                else:
                    attr = self.attr_normal
            else:
                attr = self.attr_normal
            attr.IncRef()
            return attr
        else:
            return SchGridTableBase.GetAttr(self, row, col, kind)

    def GetNumberCols(self):
        return len(self.GetColNames()) - 1

    def IsEmptyCell(self, row, col):
        rec = self.get_rec(row)
        if rec != None:
            if rec[col] == None:
                return True
            else:
                return False
        else:
            return True

    def GetValue(self, row, col):
        if row < self.rec_count:
            if row in self.rec_to_update:
                rec = (self.rec_to_update)[row]
            else:
                rec = self.get_rec(row)
            if rec != None:
                if rec[col + 1] == None:
                    return ""
                else:
                    ret = rec[col + 1]
                    if ret.__class__ == tuple or ret.__class__ == list:
                        ret = ret[0]
                    if type(ret) == str:
                        return ret
                    else:
                        return str(ret)
            else:
                return ""
        else:
            if row in self.rec_to_instert:
                return str((self.rec_to_instert)[row][col + 1])
            else:
                return ""

    def SetValue(self, row, col, value):
        if self.read_only:
            return

        if row >= self.rec_count:
            if row not in self.rec_to_instert:
                (self.rec_to_instert)[row] = copy.copy(self.default_rec)

            (self.rec_to_instert)[row][col + 1] = value

            if row == self.GetNumberRows() - 1:
                self.append_count = self.append_count + 1
                msg = wx.grid.GridTableMessage(
                    self, wx.grid.GRIDTABLE_NOTIFY_ROWS_APPENDED, 1
                )
                self.GetView().ProcessTableMessage(msg)
        else:
            if row not in self.rec_to_update:
                (self.rec_to_update)[row] = self.get_rec(row)
            (self.rec_to_update)[row][col + 1] = value
            if row in self.rec_to_delete:
                self.rec_to_delete.remove(row)

    def CanGetValueAs(self, row, col, typeName):
        colType = self.GetColTypes()[col + 1].split(":")[0]
        if typeName == colType:
            return True
        else:
            return False

    def CanSetValueAs(self, row, col, typeName):
        return self.CanGetValueAs(row, col, typeName)

    def GetColLabelValue(self, col):
        return self.GetColNames()[col + 1]

    def GetValueAsBool(self, row, col):
        colType = self.GetColTypes()[col + 1].split(":")[0]
        if colType == "bool":
            ret = self.GetValue(row, col)
            if ret:
                return True
            else:
                return False
        else:
            return False

    def SetValueAsBool(self, row, col, value):
        self.SetValue(row, col, value)

    def GetTypeName(self, row, col):
        try:
            return self.GetColTypes()[col + 1]
        except:
            print("GetTypeName:", col, self.GetColTypes())

    def DeleteRows(self, row, il):
        if self.read_only:
            return
        if row < self.rec_count:
            if row in self.rec_to_update:
                del (self.rec_to_update)[row]
            if row in self.rec_to_delete:
                self.rec_to_delete.remove(row)
            else:
                self.rec_to_delete.append(row)
            self.GetView().ForceRefresh()
            return True
        else:
            if row in self.rec_to_instert:
                msg = wx.grid.GridTableMessage(
                    self, wx.grid.GRIDTABLE_NOTIFY_ROWS_DELETED, row, il
                )
                del (self.rec_to_instert)[row]
                self.append_count = self.append_count - 1
                self.GetView().ProcessTableMessage(msg)
                self.GetView().ForceRefresh()

                for pos in list(self.rec_to_instert.keys()):
                    if pos > row:
                        (self.rec_to_instert)[pos - 1] = (self.rec_to_instert)[pos]
                        del (self.rec_to_instert)[pos]

                return True
            return False



import sys
import os
import traceback
import datetime
import wx
from asyncio import sleep

from django.template import Template, Context

from pytigon.pytigon_run import run


class WxAuto:
    def __init__(self, script_name, pos, size):
        self.script_name = script_name
        self.pos = pos
        self.size = size
        self.subtitle_id = 1
        self.start_time = datetime.datetime.now()
        self.time_unit = 0.5
        self.sys_time_unit = 0.5
        import pyautogui

        self.pyautogui = pyautogui

    def get_region(self):
        return (
            self.pos.x,
            self.pos.y,
            self.size.GetWidth(),
            self.size.GetHeight(),
        )

    async def auto_move_and_focus(self, window_name, set_focus=True):
        await sleep(self.sys_time_unit / 3)
        win = wx.Window.FindWindowByName(window_name)
        if win and set_focus:
            win.SetFocus()
            await sleep(self.sys_time_unit / 3)
        if not win:
            print("ERROR! ", window_name)
        pos = win.GetScreenPosition()
        size = win.GetSize()
        self.pyautogui.moveTo(
            pos[0] + int(size.GetWidth() / 2), pos[1] + int(size.GetHeight() / 2)
        )
        # pyautogui.click()
        await sleep(self.sys_time_unit / 3)
        return win

    async def auto_click(self, window_name):
        await self.auto_move_and_focus(window_name)
        self.pyautogui.click()

    async def auto_focus_on_img(self, image_name):
        await sleep(self.sys_time_unit / 2)
        location = self.pyautogui.locateOnScreen(image_name, region=self.get_region())
        x, y = self.pyautogui.center(location)
        self.pyautogui.moveTo(x, y)
        await sleep(self.sys_time_unit / 2)

    async def auto_click_on_img(self, image_name):
        await self.auto_focus_on_img(image_name)
        self.pyautogui.click()

    async def dropdown(self, delta):
        delta2 = int(delta)
        self.pyautogui.keyDown("alt")
        self.pyautogui.press("down")
        self.pyautogui.keyUp("alt")
        await sleep(1)
        if delta2 > 0:
            for i in range(0, delta2):
                self.pyautogui.press("down")
                await sleep(0.1)
        else:
            for i in range(0, -1 * delta2):
                self.pyautogui.press("up")
                await sleep(0.1)
        self.pyautogui.press("enter")
        await sleep(1)

    async def __lshift__(self, txt):
        if not txt:
            return
        delta = datetime.datetime.now() - self.start_time
        t1 = datetime.datetime(year=2000, month=1, day=1, hour=0, minute=0) + delta
        t2 = t1 + datetime.timedelta(seconds=3)
        s = "%d\n%02d:%02d:%02d,%d --> %02d:%02d:%02d,%d\n" % (
            self.subtitle_id,
            t1.hour,
            t1.minute,
            t1.second,
            t1.microsecond // 1000,
            t2.hour,
            t2.minute,
            t2.second,
            t2.microsecond // 1000,
        )
        with open(
            self.script_name + ".srt", "wt" if self.subtitle_id == 1 else "at"
        ) as f:
            f.write(s)
            f.write(txt)
            f.write("\n\n")
        self.subtitle_id += 1

    async def process(self, txt, argv=None, change_tab=None):
        if txt.startswith("@"):
            with open(txt[1:], "rt") as f:
                txt2 = f.read()
                if argv:
                    t = Template(txt2)
                    c = Context(argv)
                    txt2 = t.render(c)
        else:
            txt2 = txt
        if change_tab:
            if type(change_tab) == str:
                change_tab2 = change_tab.replace(";;", "\n").split("\n")
            else:
                change_tab2 = change_tab
            change_tab2.reverse()
            for item in change_tab2:
                if ":" in item:
                    x = item.split(":", 1)
                    txt2 = txt2.replace("$" + x[0], x[1])

        for line in txt2.split("\n"):
            l = line.strip()
            if l.startswith("."):
                await sleep(len(l) * self.time_unit)
            elif l.startswith("^"):
                x = l.split(":")
                if len(x) > 1:
                    await getattr(self, x[0][1:])(x[1])
                else:
                    await getattr(self, x[0][1:])()
            elif l.startswith("#"):
                pass
            elif l.startswith("<<"):
                x = l[2:].split(" ")
                for item in x:
                    item2 = item.strip()
                    if item2.startswith("^"):
                        self.pyautogui.keyUp(item2[1:])
                    elif item2.startswith("_"):
                        self.pyautogui.keyDown(item2[1:])
                    elif item2.startswith("-"):
                        await sleep((len(item2) - 1) * self.time_unit / 10)
                    else:
                        self.pyautogui.press(item2)
            else:
                if "<<" in l:
                    x = l.split("<<")
                    await (getattr(self, x[0].strip()) << x[1].strip())
                else:
                    await (self << l.strip())

    def __getattr__(self, item):
        if item.endswith("_png"):
            return ImgProxy(self, item.replace("__", "/").replace("_png", ".png"))
        else:
            return ControlProxy(self, item)


class ControlProxy:
    def __init__(self, wx_auto, control_name):
        self.wx_auto = wx_auto
        self.control_name = control_name

    async def __lshift__(self, txt):
        ctrl = await self.wx_auto.auto_move_and_focus(self.control_name)
        if txt == "click":
            self.wx_auto.pyautogui.click()
        else:
            if txt == "focus":
                pass
            else:
                ctrl.SetValue(txt)


class ImgProxy:
    def __init__(self, wx_auto, img_name):
        self.img = None
        self.wx_auto = wx_auto
        self.img_name = img_name

    async def __lshift__(self, txt):
        await self.wx_auto.auto_focus_on_img(self.img_name)
        self.wx_auto.pyautogui.click()


SCRIPT = sys.argv[-1]


def autoit(win):
    if win:
        pos = win.GetScreenPosition()
        size = win.GetSize()
    wx_auto = WxAuto(SCRIPT, pos, size)

    async def astart():
        try:
            await sleep(1)
            # if os.path.exists(SCRIPT + ".txt"):
            if SCRIPT.endswith(".txt"):
                await wx_auto.process("@" + SCRIPT)
            else:
                sys.path.insert(0, os.getcwd())
                scr = __import__(SCRIPT.rsplit(".")[0], globals(), locals(), [], 0)
                await scr.wxauto(wx_auto, wx_auto.pyautogui, wx)
        except:
            print(sys.exc_info()[0])
            print(traceback.print_exc())

    wx.GetApp().StartCoroutine(astart, win)


if __name__ == "__main__":
    setattr(wx, "pseudoimport", autoit)

    sys.argv = [
        __file__.replace("pytigon-gui", "pytigon").replace("wxauto.py", "ptig.py"),
        "--video=%s.avi" % SCRIPT,
        "--rpc=8090",
        "schdevtools",
        "--inspection",
    ]

    run()



"""Module contains SchBaseCtrl class"""

import wx

from pytigon_lib.schhtml.htmlviewer import tdata_from_html

try:
    from urllib.parse import unquote
except:
    from urllib import unquote
from pytigon_lib.schparser.html_parsers import TreeParser


class SchBaseCtrl(object):
    """Base class for all widgets"""

    def __init__(self, parent, kwds):
        """Constructor"""
        self.unique_name = None
        self.init_base(kwds, parent)
        self.accept_focus = True
        # self.acc_tab = False
        self.acc_tab_bind = False
        form = self.get_parent_form()
        if form:
            form.signal_from_child(self, "__init__")

    def init_base(self, kwds, parent=None):
        if parent:
            self.parent = parent

        self.ldatabuf = None

        self.tag = None
        if "param" in kwds:
            if "tag" in kwds["param"]:
                self.tag = kwds["param"]["tag"]

        if "href" in kwds:
            self.href = kwds["href"]
            if self.href == None:
                self.href = ""
            del kwds["href"]
        else:
            self.href = None
        if "label" in kwds:
            self.label = kwds["label"]
            del kwds["label"]
        else:
            self.label = None
        if "id" in kwds:
            self.nr_id = kwds["id"]
            del kwds["id"]
        else:
            self.nr_id = None
        if "target" in kwds:
            self.target = kwds["target"]
            del kwds["target"]
        else:
            self.target = "_blank"
        if "value" in kwds:
            self.value = kwds["value"]
            del kwds["value"]
        else:
            self.value = None
        if "valuetype" in kwds:
            self.valuetype = kwds["valuetype"]
            del kwds["valuetype"]
        else:
            self.valuetype = "data"
        if "defaultvalue" in kwds:
            self.defaultvalue = kwds["defaultvalue"]
            del kwds["defaultvalue"]
        else:
            self.defaultvalue = None
        if "length" in kwds:
            self.length = kwds["length"]
            del kwds["length"]
        else:
            self.length = 0
        if "maxlength" in kwds:
            self.maxlength = kwds["maxlength"]
            del kwds["maxlength"]
        else:
            self.maxlength = 0
        if "readonly" in kwds:
            self.readonly = kwds["readonly"]
            del kwds["readonly"]
        else:
            self.readonly = False
        if "hidden" in kwds:
            self.hidden = True
            del kwds["hidden"]
        else:
            self.hidden = False
        if "src" in kwds:
            self.src = kwds["src"]
            del kwds["src"]
        else:
            self.src = None
        if "onload" in kwds:
            self.onload = kwds["onload"]
            del kwds["onload"]
        else:
            self.onload = None
        if "tdata" in kwds:
            self.tdatabuf = None
            self.tdata = kwds["tdata"]
            del kwds["tdata"]
        else:
            self.tdata = None
            self.tdatabuf = None
        if "ldata" in kwds:
            self.ldata = kwds["ldata"]
            del kwds["ldata"]
        else:
            self.ldata = None
        if "data" in kwds:
            self.data = unquote(kwds["data"])
            del kwds["data"]
        else:
            self.data = None
        if "param" in kwds:
            self.param = kwds["param"]
            del kwds["param"]
        else:
            self.param = None
        if "style" in kwds:
            self.style = kwds["style"]
            del kwds["style"]
        else:
            self.style = None

        ctrl_process = wx.GetApp().ctrl_process
        if self.tag in ctrl_process:
            for fun in ctrl_process[self.tag]:
                fun(self)

    def after_create(self):
        """Function called immediately after object create"""

        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down_base)
        if self.onload:
            d = {"wx": wx, "self": self}
            exec(self.onload, d)

        if hasattr(self, "__ext_init__"):
            self.__ext_init__()

        self.get_parent_form().signal_from_child(self, "after_create")

    def CanAcceptFocus(self):
        if self.IsShown() and self.IsEnabled():
            return self.accept_focus
        else:
            return False

    def CanAcceptFocusFromKeyboard(self):
        return self.CanAcceptFocus()

    def set_unique_name(self, name):
        self.unique_name = name

    def get_unique_name(self):
        return self.unique_name

    def set_acc_key_tab(self, tab):
        """Set accelerator table for widget

        Args:
            tab - list or tuple of accelerator elements. Each emement has structure: (flag, keycode, callback fun)
        """
        return self.GetParent().set_acc_key_tab(self, tab)

    def on_key_down_base(self, event):
        if event.GetKeyCode() == wx.WXK_ESCAPE:
            wx.CallAfter(self.GetParent().cancel)
            # wx.CallAfter(self.GetParent().close_with_delay)
        elif event.GetKeyCode() == wx.WXK_TAB:
            if event.ShiftDown():
                self.GetParent().Navigate(self, True)
            else:
                self.GetParent().Navigate(self, False)
        else:
            event.Skip()

    def is_ctrl_block(self):
        return self.GetParent().is_ctrl_block(self)

    def load_data_from_server(self, path):
        http = wx.GetApp().get_http(self.parent)
        response = http.get(self, str(path))
        s = response.ptr()
        return s

    def load_string_from_server(self, path):
        http = wx.GetApp().get_http(self.parent)
        response = http.get(self, str(path))
        s = response.str()
        return s

    def refresh_tdata(self, html_src=None):
        if not self.tdata:
            if self.src:
                if html_src:
                    tables = html_src
                else:
                    tables = self.load_data_from_server(self.src).decode("utf-8")

                self.tdatabuf = tdata_from_html(tables, wx.GetApp().http)

    def get_tdata(self):
        if self.tdata:
            return self.tdata
        if self.tdatabuf:
            return self.tdatabuf
        if self.src:
            self.refresh_tdata()
        return self.tdatabuf

    def refresh_ldata(self):
        if not self.ldata:
            if self.src:
                lista = self.LoadDataFromServer(self.src)
                mp = TreeParser()
                mp.feed(lista)
                mp.close()
                self.ldatabuf = mp.tree_parent[0][1]

    def get_ldata(self):
        if self.ldata:
            return self.ldata
        if self.ldatabuf:
            return self.ldatabuf
        if self.src:
            self.refreshLDATA()
        return self.ldatabuf

    def get_parent_form(self):
        """Get parent wxForm parent object"""
        parent = self.parent
        while parent != None:
            if type(parent).__name__ == "SchForm":
                return parent
            parent = parent.GetParent()
        return None

    def get_parent_page(self):
        form = self.get_parent_form()
        if form:
            return form.get_parent_page()
        else:
            return None


def handle_best_size(base_class):
    class _Derived(base_class):
        def GetBestSize(self):
            w = 0
            h = 0
            if self.param and "width" in self.param:
                try:
                    w = int(self.param["width"])
                except:
                    pass
            if self.param and "height" in self.param:
                try:
                    w = int(self.param["height"])
                except:
                    pass
            if w == 0 or h == 0:
                base_size = super().GetBestSize()
                if w == 0:
                    w = base_size[0]
                if h == 0:
                    h = base_size[1]
            return (w, h)

    return _Derived
